# future
from __future__ import annotations

# stdlib
import textwrap
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

# third party
import numpy as np
from numpy.typing import ArrayLike

# relative
from ..adp.data_subject_ledger import DataSubjectLedger
from ..adp.data_subject_list import DataSubjectArray
from .lazy_repeat_array import lazyrepeatarray
from .manager import TensorChainManager
from .passthrough import PassthroughTensor  # type: ignore

_PhiTensorRef = None


def _PhiTensor() -> Type[PassthroughTensor]:
    global _PhiTensorRef
    if _PhiTensorRef is None:
        # relative
        from .autodp.phi_tensor import PhiTensor

        _PhiTensorRef = PhiTensor
    return _PhiTensorRef


def data_subject_creation_wizard(data: Any) -> List[Any]:

    w = textwrap.TextWrapper(initial_indent="\t", subsequent_indent="\t")

    welcome_msg = "Welcome to the Data Subject Annotation Wizard!!!"

    description1 = """You've arrived here because you called Tensor.private() without passing in any data_subjects!
Since the purpose of .private() is to add metadata for the support of automatic differential
privacy budgeting, you need to describe which parts of your Tensor correspond to which
real-world data subjects (data_subjects) whose privacy you want to protect. This is the only
way the system knows, for example, that it costs twice as much privacy budget when twice
as much of your data (say, 2 rows instead of 1 row) refer to the same data subject."""

    description2 = """Entities can be people (such as a medical patient), places (such as a family's address), or
even organizations (such as a business, state, or country). If you're not sure what kind of data subject
to include, just ask yourself the question, "who am I trying to protect the privacy of?". If it's
an organization, make one data subject per organization. If it's people, make one data subject per person.
If it's a group of people who are somehow similar/linked to each other (such as a family),
make each data subject a different group. For more information on differential privacy, see OpenMined's
course on the subject: https://courses.openmined.org/"""

    description3 = """Since you didn't pass in data_subjects into .private() (or you did so incorrectly), this wizard is
going to guide you through the process of annotating your data with data_subjects."""

    description4 = """In this wizard, we're going to ask you for *unique identifiers* which refer to the data_subjects
in your data. While the unique identifiers need not be personal data (they can be random strings of letters and numbers
if you like). It is ESSENTIAL that you use the same identifier when referring to the same data subject in the
data that you never accidentally refer to two data_subjects by the same identifier. Additionally, if you plan
to do any kind of data JOIN with another dataset, it is ESSENTIAL that you are using the same unique
identifiers for data_subjects as the data you're joining with. Since these unique identifiers may be personal
information, PySyft might not be able to detect if two tensors are using different identifiers for the
same person."""

    description5 = """So, in this tutorial we're going to be asking you to specify Unique Identifiers (UIDs) for each
data subject in your data. This could be an email, street address, or any other string that identifies someone
uniquely in your data and in the data you intend to use with your data (if any)."""

    print("\t" + "=" * 69)
    print(w.fill(welcome_msg))
    print("\t" + "=" * 69)
    print()
    print(w.fill(description1))
    print()
    print(w.fill(description2))
    print()
    print(w.fill(description3))
    print()
    print(w.fill(description4))
    print()
    print(w.fill(description5))
    print()

    print("\tDo you understand, and are you ready to proceed? (yes/no)")
    print()
    consent = str(input("\t"))
    print()

    if consent == "no":
        raise Exception("User cancelled data subject creation wizard!")

    print("\tExcellent! Let's begin!")
    # print("\tYou passed in a tensor with the shape:" + str(data.shape))
    print()

    print("\t" + "-" * 69)
    print()

    print(
        w.fill("Question 1: Is this entire tensor referring to the same data subject?")
    )
    print()
    print(w.fill("Examples:"))
    print("\t - a single medical scan of one patient")
    print("\t - a single spreadsheet of proprietary statistics about a business")
    print("\t - a tensor of facts about a country")
    print()
    print(
        w.fill(
            """(if the tensor is about one data subject, but it also contains multiple other data_subjects within,
such as a tensor about all the customers of one business, ask yourself, are you trying to
protect the people or the business)"""
        )
    )
    print()
    print(
        w.fill(
            "If yes, write the UID of the data subject this data is about, otherwise write 'no' "
            " because this data is about more than one data subject."
        )
    )
    print()
    single_uid = input("\t")
    print()
    if single_uid != "no":
        print("\t" + "-" * 69)
        print()
        print(
            w.fill(
                "Excellent! Your data will be annotated as referring to:"
                + str(single_uid)
            )
        )
        print()
        print(
            w.fill(
                "Congratulations! You're all done with the Data Subject Annotation Wizard!!!"
                "In the future, you can accomplish this without the wizard by running:"
            )
        )
        print()
        print(w.fill("\t.private(data_subjects='" + str(single_uid) + "')"))
        print()
        print("\t" + "=" * 69)
        return [single_uid]

    print("\t" + "-" * 69)
    print()
    print(
        w.fill(
            "Question 2: Does each row correspond to an data subject, perhaps with occasional repeats (yes/no)?"
        )
    )
    print()
    answer = str(input("\t"))
    print()
    print("\t" + "-" * 69)
    print()
    if answer == "yes":
        print(
            w.fill(
                "Question 3: Excellent! Well, since your dataset has "
                + str(data.shape[0])
                + " rows, "
                + "would you like to hand enter an data subject for each one (yes) or if there are too "
                + "many for you to hand-enter, we'll print some example code for you to run (no)."
            )
        )

        print()

        answer = str(input("\t"))

        if answer == "yes":

            print()

            data_subjects = list()
            for i in range(len(data)):
                print("\t\t" + "-" * 61)
                print()
                print(w.fill("\tData Row " + str(i) + ":" + str(data[i])))
                ent = input("\t\t What data subject is this row about:")
                data_subjects.append(ent)
                print()
            print("\t\t" + "-" * 61)
            print()
            print(
                w.fill(
                    "All done! Next time if you want to skip the wizard, call .private() like this:"
                )
            )
            print()
            print(
                w.fill(
                    ".private(data_subjects=['"
                    + data_subjects[0]
                    + "', '"
                    + data_subjects[1]
                    + "', '"
                    + data_subjects[-1]
                    + "'])"
                )
            )
            print()
            print(
                w.fill(
                    " where you pass in data_subjects as a list of strings, one per row. As long as you"
                    " pass in the same number of data_subjects as there are rows in your tensor, it will"
                    " automatically detect you have and assume you mean one data subject per row."
                )
            )
            return data_subjects

        elif answer == "no":

            print()

            print(
                w.fill(
                    "Excellent. Well, in that case you'll need to re-run .private() but pass in"
                    " a list of strings where each string is a unique identifier for an data subject, and where"
                    " the length of the list is equal to the number of rows in your tensor. Like so:"
                )
            )

            print()
            print(w.fill(".private(data_subjects=['bob', 'alice', 'john'])"))
            print()
            print(
                " Now just to make sure I don't corrupt your tensor - I'm going to throw an exception."
            )
            print()
            raise Exception(
                "Wizard aborted. Please run .private(data_subjects=<your data_subjects>)"
                " again with your list of data subject unique identifiers (strings),"
                "one per row of your tensor."
            )
    elif answer == "no":

        print(
            w.fill(
                "Question 3: Is your data one data subject for every column (yes/no)?"
            )
        )

        print()

        answer = str(input("\t"))

        print()

        if answer == "yes":
            print(
                w.fill(
                    "We don't yet support this form of injestion. Please transpose your data"
                    " into one data subject per row and re-run the wizard. Aborting:)"
                )
            )

            raise Exception("Wizard aborted.")

        elif answer == "no":

            print(
                w.fill(
                    "It sounds like your tensor is a random assortment of data_subjects (and perhaps empty/non"
                    "-data_subjects). If you have empty values, just create random data_subjects for them for now. If "
                    "you have various data_subjects scattered throughout your tensor (not organized by row), then "
                    "you'll need to pass in a np.ndarray of strings which is identically shaped to your data in data "
                    "subjects like so:"
                )
            )

            print()
            print("\t\ttensor = sy.Tensor(np.ones((2,2)).astype(np.int32))")
            print()
            print(
                "\t\tdata_subjects = np.array([['bob', 'alice'],['charlie', 'danielle']])"
            )
            print()
            print(
                "\t\ttensor.private(min_vals=0, max_vals=1, data_subjects=data_subjects))"
            )
            print()
            print(
                "Aborting wizard now so that you rcan re-run .private with the right parameters."
            )
            print()
            raise Exception(
                "Wizard aborted. Please run .private(data_subjects=<your data_subjects>)"
                " again with your np.ndarray of data subject unique identifiers (strings),"
                " one per value of your tensor and where your np.ndarray of data_subjects is"
                " the same shape as your data."
            )
    print()

    print("\t" + "_" * 69)
    raise Exception(
        "Not sure what happened... this code shouldn't have been reached. Try answering questions with "
        "options given by the prompts (such as yes/no)."
    )


class PhiTensorAncestor(TensorChainManager):
    """Inherited by any class which might have or like to have SingleEntityPhiTensor in its chain
    of .child objects"""

    def __init__(self, child: Any) -> None:
        self.child = child

    @property
    def shape(self) -> Tuple[Any, ...]:
        return self.child.shape

    @property
    def min_vals(self):  # type: ignore
        return self.__class__(self.child.min_vals)

    @property
    def max_vals(self):  # type: ignore
        return self.__class__(self.child.max_vals)

    @property
    def gamma(self):  # type: ignore
        return self.__class__(self.child.gamma)

    def publish(
        self,
        get_budget_for_user: Callable,
        deduct_epsilon_for_user: Callable,
        ledger: DataSubjectLedger,
        sigma: float,
    ) -> Any:
        return self.child.publish(
            get_budget_for_user, deduct_epsilon_for_user, ledger, sigma
        )

    def copy(self) -> PhiTensorAncestor:
        """This should certainly be implemented by the subclass but adding this here to
        satisfy mypy."""

        return NotImplemented

    def annotated_with_dp_metadata(
        self,
        min_val: ArrayLike,
        max_val: ArrayLike,
        data_subjects: Optional[Any] = None,
        skip_blocking_checks: bool = False,
    ) -> PhiTensorAncestor:
        print("Tensor annotated with DP Metadata")
        return self.private(
            min_val=min_val,
            max_val=max_val,
            data_subjects=data_subjects,
            skip_blocking_checks=skip_blocking_checks,
        )

    def private(
        self,
        min_val: ArrayLike,
        max_val: ArrayLike,
        data_subjects: Optional[Any] = None,
        skip_blocking_checks: bool = False,
    ) -> PhiTensorAncestor:
        return self.copy()._private(
            min_val=min_val,
            max_val=max_val,
            data_subjects=data_subjects,
            skip_blocking_checks=skip_blocking_checks,
        )

    def _private(
        self,
        min_val: ArrayLike,
        max_val: ArrayLike,
        data_subjects: Optional[Any] = None,
        skip_blocking_checks: bool = False,
    ) -> PhiTensorAncestor:
        # PHASE 1: RUN CHECKS

        # Check 1: Is self.child a compatible type? We only support DP and SMPC for a few types.
        if not isinstance(self.child, np.ndarray):

            msg = (
                "At present, you can only call .private() "
                + "on syft.Tensor objects wrapping numpy arrays. You called it on a "
                + "syft.Tensor wrapping a "
                + str(type(self.child))
            )

            if isinstance(self.child, np.ndarray):
                msg += " with dtype:" + str(getattr(self.child, "dtype", None))

            raise TypeError(msg)

        # Check 2: If data_subjects == None, then run the entity creation tutorial
        if data_subjects is None:
            if skip_blocking_checks:
                raise Exception(
                    "Error: 'data_subjects' argument to .private() must not be None!"
                )
            print(
                "ALERT: You didn't pass in any data_subjects. Launching data subject wizard...\n"
            )
            data_subjects = data_subject_creation_wizard(self.child)

        # Check 3: If data_subjects is a string, make it a list with one entity in it
        if isinstance(data_subjects, str):
            data_subjects = [DataSubjectArray(data_subjects)]
        elif isinstance(data_subjects, DataSubjectArray):
            data_subjects = [data_subjects]
        # Check 4: If data_subjects are a list, are the items strings or DataSubjectArray objects.
        # If they're strings lets create DataSubjectArray objects.

        # if isinstance(data_subjects, (list, tuple)):
        #     data_subjects = np.array(data_subjects)

        # if len(data_subjects) != 1 and data_subjects.shape != self.shape:
        #     raise Exception(
        #         "Entities shape doesn't match data shape. If you're"
        #         " going to pass in something other than 1 entity for the"
        #         " entire tensor or one entity per row, you're going to need"
        #         " to make the np.ndarray of data_subjects have the same shape as"
        #         " the tensor you're calling .private() on. Try again."
        #     )

        if not isinstance(data_subjects, DataSubjectArray):
            data_subjects = DataSubjectArray.from_objs(data_subjects)

        # SKIP check temporarily
        # for entity in one_hot_lookup:
        #     if not isinstance(entity, (np.integer, str, DataSubject)):
        #         raise ValueError(
        #             f"Expected DataSubject to be either string or DataSubject object, but type is {type(entity)}"
        #         )
        if data_subjects.shape != self.shape:
            raise ValueError(
                f"DataSubjects shape: {data_subjects.shape} should match data shape: {self.shape}"
            )

        if isinstance(min_val, (bool, int, float)):
            min_vals = np.array(min_val).ravel()  # make it 1D
        else:
            raise Exception(
                "min_vals should be either float,int,bool got "
                + str(type(min_val))
                + " instead."
            )

        if isinstance(max_val, (bool, int, float)):
            max_vals = np.array(max_val).ravel()  # make it 1D
        else:
            raise Exception(
                "min_vals should be either float,int,bool got "
                + str(type(max_val))
                + " instead."
            )

        if min_vals.shape != self.child.shape:
            min_vals = lazyrepeatarray(min_vals, self.child.shape)

        if max_vals.shape != self.child.shape:
            max_vals = lazyrepeatarray(max_vals, self.child.shape)

        self.replace_abstraction_top(
            tensor_type=_PhiTensor(),
            child=self.child,
            min_vals=min_vals,
            max_vals=max_vals,
            data_subjects=data_subjects,  # type: ignore
        )  # type: ignore

        return self



# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

# third party
from nacl.signing import VerifyKey

# relative
from ..... import lib
from ....common.serde.serializable import serializable
from ....common.uid import UID
from ....io.address import Address
from ....store.storeable_object import StorableObject
from ....tensor.smpc.share_tensor import ShareTensor
from ....tensor.smpc.utils import ispointer
from ...abstract.node import AbstractNode
from .common import ImmediateActionWithoutReply
from .greenlets_switch import retrieve_object


@serializable(recursive_serde=True)
class PRZSAction(ImmediateActionWithoutReply):
    """
    When executing a PRZSAction, a :class:`Node` will run a ShareTensor przs method
    and keep the returned value in its store.

    Attributes:
         args: args to pass to the function. They should be pointers to objects
            located on the :class:`Node` that will execute the action.
         kwargs: kwargs to pass to the function. They should be pointers to objects
            located on the :class:`Node` that will execute the action
        is_dp_tensor: Set to True, if the tensor chain contains a DP Tensor
    """

    __attr_allowlist__ = [
        "path",
        "args",
        "kwargs",
        "id_at_location",
        "is_dp_tensor",
        "address",
        "msg_id",
        "_id",
    ]

    def __init__(
        self,
        path: str,
        args: List[Any],
        kwargs: Dict[Any, Any],
        id_at_location: UID,
        is_dp_tensor: bool,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        self.path = path
        self.args = args
        self.kwargs = kwargs
        self.id_at_location = id_at_location
        self.is_dp_tensor = is_dp_tensor
        # logging needs .path to exist before calling
        # this which is why i've put this super().__init__ down here
        super().__init__(address=address, msg_id=msg_id)

    @property
    def pprint(self) -> str:
        return f"PRZSAction({self.path})"

    def __repr__(self) -> str:
        arg_names = ",".join([a.__class__.__name__ for a in self.args])
        kwargs_names = ",".join(
            [f"{k}={v.__class__.__name__}" for k, v in self.kwargs.items()]
        )
        return f"PRZSAction ({arg_names}, {kwargs_names})"

    def execute_action(self, node: AbstractNode, verify_key: VerifyKey) -> None:

        (
            upcasted_args,
            upcasted_kwargs,
        ) = lib.python.util.upcast_args_and_kwargs(self.args, self.kwargs)

        if "value" in upcasted_kwargs and ispointer(upcasted_kwargs["value"]):
            upcasted_kwargs["value"] = retrieve_object(
                node, upcasted_kwargs["value"].id_at_location, self.path
            ).data

        if not self.is_dp_tensor:
            result = ShareTensor.generate_przs(*upcasted_args, **upcasted_kwargs)
        else:
            result = ShareTensor.generate_przs_on_dp_tensor(
                *upcasted_args, **upcasted_kwargs
            )

        result_read_permissions = {
            node.verify_key: node.id,
            verify_key: None,  # we dont have the passed in sender's UID
        }
        result_write_permissions = {
            node.verify_key: node.id,
            verify_key: None,  # we dont have the passed in sender's UID
        }

        if not isinstance(result, StorableObject):
            store_obj = StorableObject(
                id=self.id_at_location,
                data=result,
                read_permissions=result_read_permissions,
                write_permissions=result_write_permissions,
            )

        node.store[self.id_at_location] = store_obj



"""The strategy that all GC Heuristics should implement."""
# stdlib
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List
from typing import Type

# relative
from ..pointer import Pointer


class GCStrategy(ABC):
    """The Strategy that all GC Heuristics should inherit."""

    REGISTERED_GC_STRATEGIES: Dict[str, Type["GCStrategy"]] = {}

    def __init__(self, *args: List[Any], **kwargs: Dict[Any, Any]) -> None:
        """Initialize an object for GCStratgy.
        The subclasses that require more logic to be used, should implement
        the __init__ method
        """
        pass

    @staticmethod
    def _register(cls: Type["GCStrategy"]) -> None:
        """Add a concrete implementation for the GC Strategy to a dictionary
        of known strategies.

        This is used to have a simpler way when we initialize the GarbageCollection
        class.

        Args:
            cls (Type[GCStrategy]): The subclass that inherits from GCStrategy

        Return:
            None
        """
        gc_strategy_name = cls.__name__.lower()
        if gc_strategy_name in GCStrategy.REGISTERED_GC_STRATEGIES:
            raise ValueError(f"{gc_strategy_name} already registered!")

        GCStrategy.REGISTERED_GC_STRATEGIES[gc_strategy_name] = cls

    def __init_subclass__(cls: Type["GCStrategy"]) -> None:
        """Called when a class inherits the GCStrategy.

        Args:
            cls (Type[GCStrategy]): The subclass that inherits from GCStrategy

        Return:
            None
        """
        super().__init_subclass__()
        GCStrategy._register(cls)

    @abstractmethod
    def reap(self, pointer: Pointer) -> None:
        """What happens when a Ponter is deleted.
        It should be implemented in the GCStrategy that extends this class
        """
        pass



class PyGridClientException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class RequestAPIException(Exception):
    def __init__(self, message: str) -> None:
        message = "Something went wrong during this request: " + message
        super().__init__(message)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/action/get_object.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.core.store import (
    store_object_pb2 as proto_dot_core_dot_store_dot_store__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n.proto/core/node/common/action/get_object.proto\x12\x1csyft.core.node.common.action\x1a%proto/core/common/common_object.proto\x1a#proto/core/store/store_object.proto\x1a\x1bproto/core/io/address.proto"\xcc\x01\n\x0fGetObjectAction\x12-\n\x0eid_at_location\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12%\n\x06msg_id\x18\x02 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\ndelete_obj\x18\x05 \x01(\x08"\x97\x01\n\x18GetObjectResponseMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12,\n\x03obj\x18\x02 \x01(\x0b\x32\x1f.syft.core.store.StorableObject\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.core.node.common.action.get_object_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _GETOBJECTACTION._serialized_start = 186
    _GETOBJECTACTION._serialized_end = 390
    _GETOBJECTRESPONSEMESSAGE._serialized_start = 393
    _GETOBJECTRESPONSEMESSAGE._serialized_end = 544
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey
from typing_extensions import final

# relative
from .....common.serde.serializable import serializable
from .....common.uid import UID
from ....domain_interface import DomainInterface
from ....domain_msg_registry import DomainMessageRegistry
from ....registry import VMMessageRegistry
from ...permissions.permissions import BasePermission
from ...permissions.user_permissions import NoRestriction
from ..generic_payload.syft_message import NewSyftMessage as SyftMessage
from ..generic_payload.syft_message import ReplyPayload
from ..generic_payload.syft_message import RequestPayload


@serializable(recursive_serde=True)
@final
class ObjectDeleteMessage(SyftMessage, DomainMessageRegistry, VMMessageRegistry):

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        """Payload fields and types used during a User Creation Request."""

        id_at_location: str

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        """Payload fields and types used during a User Creation Response."""

        message: str = "Deleted Successfully."

    request_payload_type = (
        Request  # Converts generic syft dict into a ObjectDeleteMessage Request object.
    )
    reply_payload_type = (
        Reply  # Creates a proper Reply payload message structure as a response.
    )

    def run(  # type: ignore
        self, node: DomainInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:  # type: ignore

        # TODO: We can have it run async and have a cron/periodic task to clean up failed deletes.

        # relative
        from ......logger import critical
        from ......logger import debug

        try:
            debug(
                f"Calling delete on Object with ID {self.payload.id_at_location} in store."  # ignore
            )
            id_at_location = UID.from_string(self.payload.id_at_location)
            if not node.store.is_dataset(key=id_at_location):  # type: ignore
                node.store.delete(key=id_at_location)
        except Exception as e:
            log = f"> ObjectDeleteMessage delete exception {self.payload.id_at_location} {e}"
            critical(log)

        return ObjectDeleteMessage.Reply()

    def get_permissions(self) -> List[Type[BasePermission]]:
        """Returns the list of permission classes."""
        # Needs to be replaced with UserHasWritePermissionToData once who has permissions to delete is sorted
        return [NoRestriction]



# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple


class IndexableTrait(type):
    """
    IndexableTrait is used as a metaclass for our UnionGenerator to be able to
    set __getitem__ on the class to enable the UnionGenerator[...] syntax.
    """

    def __getitem__(self, union_types: Tuple[str, ...]) -> str:
        return UnionGenerator.from_qualnames(list(union_types))


class UnionGenerator(type, metaclass=IndexableTrait):
    """
    UnionGenerator is a metaclass used to generate on the spot union types
    from their qualnames.

    Eg. UnionGenerator["syft.lib.python.Int", "syft.lib.python.Float"] will generate
    in the current module the FloatIntUnion type that will have no functions added
    to the type itself until the AST is finished to be generated. If new types of
    unions have to be generated afterwards:
    1. you might be using the ast in a wrong way.
    2. if you know what you are doing, you have to regenerate the ast.
    """

    SUFFIX_UNION = "Union"

    def __new__(cls, name: str, bases: tuple, dct: dict) -> Any:
        dct["__qualname__"] = "syft.lib.misc.union." + name
        dct["__name__"] = name
        new_type = super().__new__(cls, name, bases, dct)
        # insert our new union type in the global scope of the module
        globals()[new_type.__name__] = new_type
        return new_type

    @staticmethod
    def from_qualnames(union_types: List[str]) -> str:
        # sorting to not generate the same permutation of the types
        union_types = sorted(union_types)

        name = "".join([union_type.split(".")[-1] for union_type in union_types])
        name += UnionGenerator.SUFFIX_UNION

        # searching in the cache to not solve the union type multiple times
        if name in union_cache:
            return name

        target_type = UnionGenerator(name, tuple(), {})

        # adding the type and it's union types to a lazy dict that will be solved
        # when the ast is created.
        lazy_pairing[target_type] = union_types
        qualname = target_type.__qualname__
        union_cache.add(qualname)
        return qualname


# store here already solved union types
union_cache: Set[str] = set()

# store here union types to be solved after the AST has been created and we
# can generate our misc types based on them.
lazy_pairing: Dict[UnionGenerator, List[str]] = {}



# future
from __future__ import annotations

# stdlib
from typing import Optional
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # relative
    from .... import Tensor

SMPC_CONTEXT: dict = {}
FPT_CONTEXT: dict = {}
tensor_values: Optional[Tensor] = None



# stdlib
from typing import Any
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey
from typing_extensions import final

# relative
from ...... import serialize
from ......proto.core.node.common.service.simple_service_pb2 import (
    SimpleMessage as SimpleMessage_PB,
)
from ......proto.core.node.common.service.simple_service_pb2 import (
    SimpleReplyMessage as SimpleReplyMessage_PB,
)
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.deserialize import _deserialize
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address
from ....abstract.node import AbstractNode


@serializable(recursive_serde=True)
class NodeRunnableMessageWithReply:

    __attr_allowlist__ = ["stuff"]

    def __init__(self, stuff: str) -> None:
        self.stuff = stuff

    def run(self, node: AbstractNode, verify_key: Optional[VerifyKey] = None) -> Any:
        return (
            "Nothing to see here..." + self.stuff
        )  # leaving this in for the test suite

    def prepare(self, address: Address, reply_to: Address) -> "SimpleMessage":
        return SimpleMessage(address=address, reply_to=reply_to, payload=self)


@serializable()
@final
class SimpleMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        payload: NodeRunnableMessageWithReply,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.payload = payload

    def _object2proto(self) -> SimpleMessage_PB:
        return SimpleMessage_PB(
            payload=serialize(self.payload, to_bytes=True),
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(proto: SimpleMessage_PB) -> "SimpleMessage":
        return SimpleMessage(
            payload=_deserialize(blob=proto.payload, from_bytes=True),
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return SimpleMessage_PB


@serializable()
class SimpleReplyMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        payload: NodeRunnableMessageWithReply,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.payload = payload

    def _object2proto(self) -> SimpleReplyMessage_PB:
        return SimpleReplyMessage_PB(
            payload=serialize(self.payload, to_bytes=True),
            msg_id=serialize(self.id),
            address=serialize(self.address),
        )

    @staticmethod
    def _proto2object(proto: SimpleReplyMessage_PB) -> "SimpleReplyMessage":
        return SimpleReplyMessage(
            payload=_deserialize(proto.payload, from_bytes=True),
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return SimpleReplyMessage_PB



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/action/smpc_action_message.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n7proto/core/node/common/action/smpc_action_message.proto\x12\x1csyft.core.node.common.action\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\x8c\x04\n\x11SMPCActionMessage\x12\x13\n\x0bname_action\x18\x01 \x01(\t\x12&\n\x07self_id\x18\x02 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61rgs_id\x18\x03 \x03(\x0b\x32\x15.syft.core.common.UID\x12P\n\tkwargs_id\x18\x04 \x03(\x0b\x32=.syft.core.node.common.action.SMPCActionMessage.KwargsIdEntry\x12K\n\x06kwargs\x18\x05 \x03(\x0b\x32;.syft.core.node.common.action.SMPCActionMessage.KwargsEntry\x12-\n\x0eid_at_location\x18\x06 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x07 \x01(\x0b\x32\x15.syft.core.io.Address\x12%\n\x06msg_id\x18\x08 \x01(\x0b\x32\x15.syft.core.common.UID\x1a\x46\n\rKwargsIdEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12$\n\x05value\x18\x02 \x01(\x0b\x32\x15.syft.core.common.UID:\x02\x38\x01\x1a-\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.core.node.common.action.smpc_action_message_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _SMPCACTIONMESSAGE_KWARGSIDENTRY._options = None
    _SMPCACTIONMESSAGE_KWARGSIDENTRY._serialized_options = b"8\001"
    _SMPCACTIONMESSAGE_KWARGSENTRY._options = None
    _SMPCACTIONMESSAGE_KWARGSENTRY._serialized_options = b"8\001"
    _SMPCACTIONMESSAGE._serialized_start = 158
    _SMPCACTIONMESSAGE._serialized_end = 682
    _SMPCACTIONMESSAGE_KWARGSIDENTRY._serialized_start = 565
    _SMPCACTIONMESSAGE_KWARGSIDENTRY._serialized_end = 635
    _SMPCACTIONMESSAGE_KWARGSENTRY._serialized_start = 637
    _SMPCACTIONMESSAGE_KWARGSENTRY._serialized_end = 682
# @@protoc_insertion_point(module_scope)



"""
This source file aims to replace the standard slice object/function provided by Python
to be handled by the PySyft's abstract syntax tree data structure during a remote call.
"""

# stdlib
from typing import Any
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# relative
from ...core.common import UID
from ...core.common.serde.serializable import serializable
from ...proto.lib.python.slice_pb2 import Slice as Slice_PB
from .primitive_factory import PrimitiveFactory
from .primitive_interface import PyPrimitive
from .types import SyPrimitiveRet
from .util import upcast


@serializable()  # This decorator turns this class serializable.
class Slice(PyPrimitive):
    def __init__(
        self,
        start: Any = None,
        stop: Optional[Any] = None,
        step: Optional[Any] = None,
        id: Optional[UID] = None,
    ):
        """
        This class will receive start, stop, step and ID as valid parameters.

        Args:
            start (Any): Index/position where the slicing of the object starts.
            stop (Any): Index/position which the slicing takes place. The slicing stops at index stop-1.
            step (Any): Determines the increment between each index for slicing.
            id (UID): PySyft's objects have an unique ID related to them.
        """

        # first, second, third
        if stop is None and step is None:
            # slice treats 1 arg as stop not start
            stop = start
            start = None

        self.value = slice(start, stop, step)
        self._id: UID = id if id else UID()

    @property
    def id(self) -> UID:
        """
        We reveal PyPrimitive.id as a property to discourage users and
        developers of Syft from modifying .id attributes after an object
        has been initialized.

        Returns:
            UID: The unique ID of the object.
        """
        return self._id

    def __eq__(self, other: Any) -> SyPrimitiveRet:
        """
        Compare if self == other.

        Args:
            other (Any): Object to be compared.
        Returns:
            SyPrimitiveRet: returns a PySyft boolean format checking if self == other.
        """
        res = self.value.__eq__(upcast(other))
        return PrimitiveFactory.generate_primitive(value=res)

    def __ge__(self, other: Any) -> SyPrimitiveRet:
        """
        Compare if self >= other.

        Args:
            other (Any): Object to be compared.
        Returns:
            SyPrimitiveRet: returns a PySyft boolean format checking if self >= other.
        """
        res = self.value.__ge__(upcast(other))  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def __gt__(self, other: Any) -> SyPrimitiveRet:
        """
        Compare if self > other.

        Args:
            other (Any): Object to be compared.
        Returns:
            SyPrimitiveRet: returns a PySyft boolean format checking if self > other.
        """
        res = self.value.__gt__(upcast(other))  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def __le__(self, other: Any) -> SyPrimitiveRet:
        """
        Compare if self <= other.

        Args:
            other (Any): Object to be compared.
        Returns:
            SyPrimitiveRet: returns a PySyft boolean format checking if self <= other.
        """
        res = self.value.__le__(upcast(other))  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def __lt__(self, other: Any) -> SyPrimitiveRet:
        """
        Compare if self < other.

        Args:
            other (Any): Object to be compared.
        Returns:
            SyPrimitiveRet: returns a PySyft boolean format checking if self < other.
        """
        res = self.value.__lt__(upcast(other))  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def __ne__(self, other: Any) -> SyPrimitiveRet:
        """
        Compare if self != other.

        Args:
            other (Any): Object to be compared.
        Returns:
            SyPrimitiveRet: returns a PySyft boolean format checking if self != other.
        """
        res = self.value.__ne__(upcast(other))
        return PrimitiveFactory.generate_primitive(value=res)

    def __str__(self) -> str:
        """Slice's string representation

        Returns:
            str: The string representation of this Slice object.
        """
        return self.value.__str__()

    def indices(self, index: int) -> tuple:
        """
        Assuming a sequence of length len, calculate the start and stop
        indices, and the stride length of the extended slice described by
        the Slice object. Out of bounds indices are clipped in
        a manner consistent with the handling of normal slices.

        Args:
            index (int): Input index.
        Returns:
            Tuple: A tuple of concrete indices for a range of length len.
        """
        res = self.value.indices(index)
        return PrimitiveFactory.generate_primitive(value=res)

    @property
    def start(self) -> Optional[int]:
        """
        Index/position where the slicing of the object starts.

        Returns:
            int: Index where the slicing starts.
        """
        return self.value.start

    @property
    def step(self) -> Optional[int]:
        """
        Increment between each index for slicing.

        Returns:
            int: Slices' increment value.
        """
        return self.value.step

    @property
    def stop(self) -> Optional[int]:
        """
        Index/position which the slicing takes place.

        Returns:
            int: Slices' stop value.
        """
        return self.value.stop

    def upcast(self) -> slice:
        """
        Returns the standard python slice object.

        Returns:
            slice: returns a default python slice object represented by this object instance.
        """
        return self.value

    def _object2proto(self) -> Slice_PB:
        """
        Serialize  the Slice object instance returning a protobuf.

        Returns:
            Slice_PB: returns a protobuf object class representing this Slice object.
        """
        slice_pb = Slice_PB()
        if self.start:
            slice_pb.start = self.start
            slice_pb.has_start = True

        if self.stop:
            slice_pb.stop = self.stop
            slice_pb.has_stop = True

        if self.step:
            slice_pb.step = self.step
            slice_pb.has_step = True

        slice_pb.id.CopyFrom(sy.serialize(obj=self._id))

        return slice_pb

    @staticmethod
    def _proto2object(proto: Slice_PB) -> "Slice":
        """
        Deserialize a protobuf object creating a new Slice object instance.

        Args:
            proto (Slice_PB): Protobuf object representing a serialized slice object.
        Returns:
            Slice: PySyft Slice object instance.
        """
        id_: UID = sy.deserialize(blob=proto.id)
        start = None
        stop = None
        step = None
        if proto.has_start:
            start = proto.start

        if proto.has_stop:
            stop = proto.stop

        if proto.has_step:
            step = proto.step

        return Slice(
            start=start,
            stop=stop,
            step=step,
            id=id_,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """
        Returns the proper Slice protobuf schema.

        Returns:
            Slice_PB: Returns the Slice's Protobuf class definition.
        """
        return Slice_PB



# stdlib
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import numpy as np
from numpy.typing import NDArray

# relative
from ....common.serde.serializable import serializable
from ...autodp.gamma_tensor import GammaTensor
from ...autodp.phi_tensor import PhiTensor
from ..initializations import XavierInitialization
from .base import Layer


@serializable(recursive_serde=True)
class Linear(Layer):
    __attr_allowlist__ = (
        "n_out",
        "n_in",
        "out_shape",
        "W",
        "b",
        "dW",
        "db",
        "last_input",
        "init",
        "input_shape",
        "out_shape",
    )

    def __init__(self, n_out: int, n_in: Optional[int] = None) -> None:
        self.n_out = n_out
        self.n_in = n_in
        self.out_shape = (n_out,)

        self.W: Optional[Union[NDArray, PhiTensor, GammaTensor]] = None
        self.b: Optional[Union[NDArray, PhiTensor, GammaTensor]] = None
        self.dW: Optional[Union[NDArray, PhiTensor, GammaTensor]] = None
        self.db: Optional[Union[NDArray, PhiTensor, GammaTensor]] = None
        self.last_input: Optional[Union[PhiTensor, GammaTensor]] = None
        self.init = XavierInitialization()

    def connect_to(self, prev_layer: Optional[Layer] = None) -> None:
        if prev_layer is None:
            if self.n_in is None:
                raise ValueError(
                    "`self.n_in` cannot be None since this is the first layer."
                )
            n_in = self.n_in
        else:
            if len(prev_layer.out_shape) != 2:  # type: ignore
                raise ValueError(
                    f"The out_shape should be of length 2. \
                    Current shape is : {len(prev_layer.out_shape)}"  # type: ignore
                )

            n_in = int(prev_layer.out_shape[-1])  # type: ignore

        self.W = self.init((n_in, self.n_out))
        self.b = np.zeros((self.n_out,))
        self.input_shape = (n_in,)
        self.out_shape = (self.n_out,)

    def forward(
        self,
        input: Union[PhiTensor, GammaTensor],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> Union[PhiTensor, GammaTensor]:
        self.last_input = input

        if self.W is None:
            raise ValueError(
                "`self.W` is None. Please check if self.W is correctly initialized."
            )

        return input.dot(self.W) + self.b

    def backward(
        self,
        pre_grad: Union[PhiTensor, GammaTensor],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> Optional[Union[PhiTensor, GammaTensor]]:

        if self.last_input is None:
            raise ValueError(
                "last_input to the layer is None. \
                Please verify if the `self.last_input` is correctly initialized."
            )

        if self.W is None:
            raise ValueError(
                "self.W is None. \
                Please verify if the `self.W` is correctly initialized."
            )

        self.dW = self.last_input.T.dot(pre_grad)
        self.db = pre_grad.mean(axis=0)
        if not self.first_layer:
            return pre_grad.dot(self.W.T)

        return None

    @property
    def params(self) -> Tuple[Union[PhiTensor, GammaTensor, NDArray, None], ...]:
        return self.W, self.b

    @params.setter
    def params(
        self, new_params: Tuple[Union[PhiTensor, GammaTensor, NDArray, None], ...]
    ) -> None:
        if len(new_params) != 2:
            raise ValueError(
                f"Expected two values. Update params has length{len(new_params)}"
            )

        self.W, self.b = new_params

    @property
    def grads(self) -> Tuple[Union[PhiTensor, GammaTensor, NDArray, None], ...]:
        return self.dW, self.db



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/io/route.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import (
    connection_pb2 as proto_dot_core_dot_io_dot_connection__pb2,
)
from syft.proto.core.io import location_pb2 as proto_dot_core_dot_io_dot_location__pb2
from syft.proto.grid.connections import (
    http_connection_pb2 as proto_dot_grid_dot_connections_dot_http__connection__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x19proto/core/io/route.proto\x12\x0csyft.core.io\x1a%proto/core/common/common_object.proto\x1a\x1cproto/core/io/location.proto\x1a\x1eproto/core/io/connection.proto\x1a,proto/grid/connections/http_connection.proto"\xfc\x01\n\tSoloRoute\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x33\n\x0b\x64\x65stination\x18\x02 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocation\x12\x43\n\x12virtual_connection\x18\x03 \x01(\x0b\x32%.syft.core.io.VirtualClientConnectionH\x00\x12\x44\n\x0fgrid_connection\x18\x04 \x01(\x0b\x32).syft.grid.connections.GridHTTPConnectionH\x00\x42\x0c\n\nconnectionb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.core.io.route_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _SOLOROUTE._serialized_start = 191
    _SOLOROUTE._serialized_end = 443
# @@protoc_insertion_point(module_scope)



# stdlib
from getpass import getpass
import json
import logging
import sys
import time
from typing import Dict
from typing import Optional
from typing import Type
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey
import requests

# syft absolute
import syft as sy

# relative
from .. import GridURL
from ...core.io.connection import ClientConnection
from ...core.io.route import SoloRoute
from ...core.node.common.client import Client
from ...core.node.domain_client import DomainClient
from ...core.node.network_client import NetworkClient
from ...util import verify_tls
from .grid_connection import GridHTTPConnection

DEFAULT_PYGRID_PORT = 80
DEFAULT_PYGRID_ADDRESS = f"http://127.0.0.1:{DEFAULT_PYGRID_PORT}"


def connect(
    url: Union[str, GridURL] = DEFAULT_PYGRID_ADDRESS,
    conn_type: Type[ClientConnection] = GridHTTPConnection,
    credentials: Optional[Dict] = None,
    user_key: Optional[SigningKey] = None,
    timeout: Optional[float] = None,
) -> Client:
    # Use Server metadata
    # to build client route
    credentials = credentials if credentials else {}
    conn = conn_type(url=GridURL.from_url(url))  # type: ignore

    # get metadata and check for https redirect so that login is sent over TLS
    metadata = conn._get_metadata(timeout=timeout)  # type: ignore

    credentials = credentials if credentials is not None else {}

    if credentials:
        metadata, _user_key = conn.login(credentials=credentials)  # type: ignore
        _user_key = SigningKey(_user_key.encode(), encoder=HexEncoder)
    else:

        if not user_key:
            _user_key = SigningKey.generate()
        else:
            _user_key = user_key

    # Check node client type based on metadata response
    client_type: Union[Type[DomainClient], Type[NetworkClient]]
    if metadata.node_type == "Domain":
        client_type = DomainClient
    else:
        client_type = NetworkClient

    (
        spec_location,
        name,
        client_id,
    ) = client_type.deserialize_client_metadata_from_node(metadata=metadata)

    # Create a new Solo Route using the selected connection type
    route = SoloRoute(destination=spec_location, connection=conn)

    kwargs = {
        "name": name,
        "routes": [route],
        "signing_key": _user_key,
        "version": metadata.version,
    }

    if client_type is NetworkClient:
        kwargs["network"] = spec_location
    elif client_type is DomainClient:
        kwargs["domain"] = spec_location
    else:
        raise NotImplementedError

    # Create a new client using the selected client type
    node = client_type(**kwargs)

    return node


def login(
    url: Optional[Union[str, GridURL]] = None,
    port: Optional[int] = None,
    email: Optional[str] = None,
    password: Optional[str] = None,
    conn_type: Type[ClientConnection] = GridHTTPConnection,
    verbose: Optional[bool] = True,
    timeout: Optional[float] = None,
    retry: Optional[int] = None,
) -> Client:

    retry = 5 if retry is None else retry  # Default to 5 retries
    timeout = 10 if timeout is None else timeout  # Default to 10 seconds

    if password == "changethis":  # nosec

        if email == "info@openmined.org":
            print(
                "WARNING: CHANGE YOUR USERNAME AND PASSWORD!!! \n\nAnyone can login as an admin to your node"
                + " right now because your password is still the default PySyft username and password!!!\n"
            )
        else:
            print(
                "WARNING: CHANGE YOUR PASSWORD!!! \n\nAnyone can login into your account"
                + " right now because your password is the default PySyft password!!!\n"
            )

    # TRASK: please keep this so that people will stop putting their passwords in notebooks.
    if password == "secret":  # nosec
        print("Welcome " + str(email) + "!")
        password = getpass(prompt="Please enter you password:")

    if port is None and not url:  # if url is used, we can ignore port
        port = int(input("Please specify the port of the domain you're logging into:"))

    # TODO: build multiple route objects and let the Client decide which one to use
    if isinstance(url, GridURL):
        grid_url = url
    elif url is None:
        grid_url = GridURL(host_or_ip="docker-host", port=port, path="/api/v1/status")
        try:
            requests.get(str(grid_url), verify=verify_tls())
        except Exception:
            grid_url.host_or_ip = "localhost"
    else:
        grid_url = GridURL(host_or_ip=url, port=port)

    grid_url = grid_url.with_path("/api/v1")

    if verbose:
        sys.stdout.write("\rConnecting to " + str(grid_url.host_or_ip) + "...")

    if email is None or password is None:
        credentials = {}
        logging.info(
            "\n\nNo email and password defined in login() - connecting as anonymous user!!!\n"
        )
    else:
        credentials = {"email": email, "password": password}

    # connecting to domain
    node = None
    timeout_btw_retries = timeout
    retry_attempt = 1

    while node is None and retry_attempt <= retry:
        try:
            node = connect(
                url=grid_url,
                credentials=credentials,
                conn_type=conn_type,
                timeout=timeout,
            )
        except requests.ConnectTimeout:
            raise requests.ConnectTimeout(
                f"Connection to node with: {url} timed out. Please try again !!!"
            )
        except requests.ConnectionError as e:
            if retry_attempt <= retry:
                print(
                    f"\nConnectionError: Retrying again.... Attempt: {retry_attempt}",
                    end="\r",
                )
                time.sleep(timeout_btw_retries)
            else:
                raise e
        retry_attempt += 1

    if node is None:
        raise requests.ConnectionError(
            f"Failed to connect to node with: {url}. Please try again !!!"
        )

    if verbose:
        # bit of fanciness
        sys.stdout.write(" done! \t Logging into")
        sys.stdout.write(" " + str(node.name) + "... ")
        if email is None or password is None:
            sys.stdout.write("as GUEST...")
        time.sleep(1)  # ok maybe too fancy... but c'mon don't you want to be fancy?
        print("done!")
    else:
        print("Logging into: ...", str(node.name), " Done...")

    if sy.__version__ != node.version:
        print(
            "\n**Warning**: The syft version on your system and the node are different."
        )
        print(
            f"Version on your system: {sy.__version__}\nVersion on the node: {node.version}"
        )
        print()

    return node


def register(
    name: Optional[str] = None,
    email: Optional[str] = None,
    password: Optional[str] = None,
    url: Optional[str] = None,
    port: Optional[int] = None,
    verbose: Optional[bool] = True,
) -> Client:
    if name is None:
        name = input("Please enter your name:")

    if email is None:
        email = input("Please enter your email:")

    if password is None:
        password = getpass("Please enter your password")

    if url is None:
        url = input("Please enter URL of domain (ex: 'localhost'):")

    if port is None:
        port = int(input("Please enter the port your domain is running on:"))

    grid_url = GridURL(host_or_ip=url, port=port)

    register_url = grid_url.url + "/api/v1/register"
    myobj = {"name": name, "email": email, "password": password}

    response = requests.post(register_url, data=json.dumps(myobj))

    if "error" not in json.loads(response.text):
        if verbose:
            print("Successfully registered! Logging in...")

        return login(
            url=grid_url, port=port, email=email, password=password, verbose=verbose
        )

    raise Exception(response.text)



# stdlib
from collections.abc import Sequence
import os
from pathlib import Path
from typing import Callable
from typing import Dict
from typing import List

# third party
import capnp

CAPNP_START_MAGIC_HEADER = "capnp:"
CAPNP_END_MAGIC_HEADER = ":capnp"
CAPNP_START_MAGIC_HEADER_BYTES = CAPNP_START_MAGIC_HEADER.encode("utf-8")
CAPNP_END_MAGIC_HEADER_BYTES = CAPNP_END_MAGIC_HEADER.encode("utf-8")
CAPNP_REGISTRY: Dict[str, Callable] = {}

CapnpModule = capnp.lib.capnp._StructModule


def get_capnp_schema(schema_file: str) -> type:
    here = os.path.dirname(__file__)
    root_dir = Path(here) / ".." / ".." / ".." / "capnp"
    capnp_path = os.path.abspath(root_dir / schema_file)
    return capnp.load(str(capnp_path))


def chunk_bytes(
    data: Sequence, field_name: str, builder: capnp.lib.capnp._DynamicStructBuilder
) -> None:
    CHUNK_SIZE = int(5.12e8)  # capnp max for a List(Data) field
    list_size = len(data) // CHUNK_SIZE + 1
    data_lst = builder.init(field_name, list_size)
    END_INDEX = CHUNK_SIZE
    for idx in range(list_size):
        START_INDEX = idx * CHUNK_SIZE
        END_INDEX = min(START_INDEX + CHUNK_SIZE, len(data))
        data_lst[idx] = data[START_INDEX:END_INDEX]


def combine_bytes(capnp_list: List[bytes]) -> bytes:
    # TODO: make sure this doesn't copy, perhaps allocate a fixed size buffer
    # and move the bytes into it as we go
    bytes_value = b""
    for value in capnp_list:
        bytes_value += value
    return bytes_value


def serde_magic_header(cls: type) -> bytes:
    return (
        CAPNP_START_MAGIC_HEADER_BYTES
        + cls.__name__.encode("utf-8")
        + CAPNP_END_MAGIC_HEADER_BYTES
    )



# stdlib
from enum import Enum


class PyGridClientEnums(str, Enum):
    ENCODING = "ISO-8859-1"


class RequestAPIFields(str, Enum):
    ADDRESS = "address"
    UID = "uid"
    POINTABLE = "pointable"
    QUERY = "query"
    CONTENT = "content"
    REPLY_TO = "reply_to"
    MESSAGE = "message"
    ERROR = "error"
    RESPONSE = "response"
    SOURCE = "source"
    TARGET = "target"
    NODE = "node"
    REQUESTED_DATE = "requested_date"
    STATUS = "status"


class AssociationRequestResponses(str, Enum):
    ACCEPT = "ACCEPTED"
    DENY = "REJECTED"
    PENDING = "PENDING"


class ResponseObjectEnum(str, Enum):
    ASSOCIATION_REQUEST = "association-request"
    DATASET = "dataset"
    GROUP = "group"
    ROLE = "role"
    USER = "user"
    WORKER = "worker"
    DATA = "data"



# stdlib
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft absolute
import syft as sy

# relative
from .....logger import debug
from .....logger import traceback_and_raise
from .....logger import warning
from .....proto.core.node.common.action.get_object_pb2 import (
    GetObjectAction as GetObjectAction_PB,
)
from .....proto.core.node.common.action.get_object_pb2 import (
    GetObjectResponseMessage as GetObjectResponseMessage_PB,
)
from .....util import get_fully_qualified_name
from .....util import validate_type
from ....common.message import ImmediateSyftMessageWithoutReply
from ....common.serde.serializable import serializable
from ....common.uid import UID
from ....io.address import Address
from ....store.storeable_object import StorableObject
from ...abstract.node import AbstractNode
from ..node_service.auth import AuthorizationException
from .common import ImmediateActionWithReply


@serializable()
class GetObjectResponseMessage(ImmediateSyftMessageWithoutReply):
    """
    GetObjectResponseMessages are the type of messages that are sent in response to a
    :class:`GetObjectAction`. They contain the object that was asked for.

    Attributes:
         obj: the object being sent back to the asker.
    """

    def __init__(
        self, obj: StorableObject, address: Address, msg_id: Optional[UID] = None
    ) -> None:
        super().__init__(address=address, msg_id=msg_id)
        self.obj = obj

    def _object2proto(self) -> GetObjectResponseMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: GetObjectResponseMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        ser = sy.serialize(self.obj)

        return GetObjectResponseMessage_PB(
            msg_id=sy.serialize(self.id),
            address=sy.serialize(self.address),
            obj=ser,
        )

    @staticmethod
    def _proto2object(proto: GetObjectResponseMessage_PB) -> "GetObjectResponseMessage":
        """Creates a ObjectWithID from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of GetObjectResponseMessage
        :rtype: GetObjectResponseMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """
        return GetObjectResponseMessage(
            obj=sy.deserialize(blob=proto.obj),
            msg_id=sy.deserialize(blob=proto.msg_id),
            address=sy.deserialize(blob=proto.address),
        )

    @property
    def data(self) -> object:
        data = self.obj.data
        try:
            # TODO: Make only for DataFrames etc
            # Issue: https://github.com/OpenMined/PySyft/issues/5322
            if get_fully_qualified_name(obj=self.obj.data) not in [
                "pandas.core.frame.DataFrame",
                "pandas.core.series.Series",
            ]:
                data.tags = self.obj.tags
                data.description = self.obj.description

        except AttributeError:
            warning(
                f"'tags' and 'description' can't be attached to {type(data)} object."
            )
        return data

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return GetObjectResponseMessage_PB


@serializable()
class GetObjectAction(ImmediateActionWithReply):
    """
    This kind of action is used when a Node wants to get an object located on another Node.

    The Node receiving this action first check that the asker does have the permission to
    fetch the object he asked for. If it's the case, a :class:`GetObjectResponseMessage`
    containing the object is sent back to the asker.

    Attributes:
         id_at_location: the pointer id of the object asked for.
    """

    def __init__(
        self,
        id_at_location: UID,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
        delete_obj: bool = True,
    ):
        # TODO: remove delete_obj flag if everthing works fine.
        self.id_at_location = id_at_location
        self.delete_obj = delete_obj

        # the logger needs self.id_at_location to be set already - so we call this later
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)

    def execute_action(
        self, node: AbstractNode, verify_key: VerifyKey
    ) -> ImmediateSyftMessageWithoutReply:
        try:
            try:
                storable_object = node.store.get(self.id_at_location, proxy_only=True)
            except Exception as e:
                log = (
                    f"Unable to Get Object with ID {self.id_at_location} from store. "
                    + f"Possible dangling Pointer. {e}"
                )

                traceback_and_raise(Exception(log))

            # if you are not the root user check if your verify_key has read_permission
            if (
                verify_key != node.root_verify_key
                and verify_key not in storable_object.read_permissions
            ):
                log = (
                    f"You do not have permission to .get() Object with ID: {self.id_at_location} on node {node.name} "
                    + "Please submit a request."
                )
                traceback_and_raise(AuthorizationException(log))

            obj = validate_type(
                storable_object.clean_copy(settings=node.settings), StorableObject
            )

            if obj.is_proxy:
                obj.data.generate_presigned_url(settings=node.settings, public_url=True)

            msg = GetObjectResponseMessage(obj=obj, address=self.reply_to, msg_id=None)

            debug(
                f"Returning Object with ID: {self.id_at_location} {type(storable_object.data)}"
            )
            return msg
        except Exception as e:
            traceback_and_raise(e)
        raise Exception(f"Unable to execute action with {type(self)}")

    @property
    def pprint(self) -> str:
        return f"GetObjectAction({self.id_at_location})"

    def _object2proto(self) -> GetObjectAction_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: ObjectWithID_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetObjectAction_PB(
            id_at_location=sy.serialize(self.id_at_location, to_proto=True),
            msg_id=sy.serialize(self.id, to_proto=True),
            address=sy.serialize(self.address, to_proto=True),
            reply_to=sy.serialize(self.reply_to, to_proto=True),
            delete_obj=self.delete_obj,
        )

    @staticmethod
    def _proto2object(proto: GetObjectAction_PB) -> "GetObjectAction":
        """Creates a ObjectWithID from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of GetObjectAction
        :rtype: GetObjectAction

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetObjectAction(
            id_at_location=sy.deserialize(blob=proto.id_at_location),
            msg_id=sy.deserialize(blob=proto.msg_id),
            address=sy.deserialize(blob=proto.address),
            reply_to=sy.deserialize(blob=proto.reply_to),
            delete_obj=proto.delete_obj,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return GetObjectAction_PB



# stdlib
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import numpy as np
from numpy.typing import NDArray

# relative
from ...common.serde.serializable import serializable
from ..autodp.gamma_tensor import GammaTensor
from ..autodp.phi_tensor import PhiTensor
from .layers.base import Layer
from .utils import dp_maximum


@serializable(recursive_serde=True)
class Optimizer:
    """Abstract optimizer base class.

    Parameters
    ----------
    clip : float
        If smaller than 0, do not apply parameter clip.
    lr : float
        The learning rate controlling the size of update steps
    decay : float
        Decay parameter for the moving average. Must lie in [0, 1) where
        lower numbers means a shorter “memory”.
    lr_min : float
        When adapting step rates, do not move below this value. Default is 0.
    lr_max : float
        When adapting step rates, do not move above this value. Default is inf.
    """

    __attr_allowlist__: Tuple[str, ...] = (
        "lr",
        "clip",
        "decay",
        "lr_min",
        "lr_max",
        "iterations",
    )

    def __init__(
        self,
        lr: float = 0.001,
        clip: int = -1,
        decay: float = 0.0,
        lr_min: float = 0.0,
        lr_max: float = np.inf,
        *args: Tuple,
        **kwargs: Dict,
    ):
        self.lr = lr
        self.clip = clip
        self.decay = decay
        self.lr_min = lr_min
        self.lr_max = lr_max

        self.iterations: int = 0

    def update(self, layers: List[Layer]) -> None:
        self.iterations += 1

        self.lr *= 1.0 / 1 + self.decay * self.iterations
        self.lr = np.clip(self.lr, self.lr_min, self.lr_max)

    def __str__(self) -> str:
        return self.__class__.__name__


@serializable(recursive_serde=True)
class Adamax(Optimizer):
    """
    Parameters
    ----------
    beta1 : float
        Exponential decay rate for the first moment estimates.
    beta2 : float
        Exponential decay rate for the second moment estimates.
    epsilon : float
        Constant for numerical stability.
    References
    ----------
    .. [1] Kingma, Diederik, and Jimmy Ba (2014):
           Adam: A Method for Stochastic Optimization.
           arXiv preprint arXiv:1412.6980.
    """

    __attr_allowlist__ = (
        "beta1",
        "beta2",
        "epsilon",
        "ms",
        "vs",
        "lr",
        "iterations",
    )

    def __init__(
        self,
        beta1: float = 0.9,
        beta2: float = 0.999,
        epsilon: float = 1e-8,
        *args: Tuple,
        **kwargs: Dict,
    ):
        super(Adamax, self).__init__(args=args, kwargs=kwargs)

        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon

        self.ms: Optional[List[NDArray]] = None
        self.vs: Optional[List[NDArray]] = None

    def update(self, layers: List[Layer]) -> None:

        # init
        self.iterations += 1
        a_t = self.lr / (1 - np.power(self.beta1, self.iterations))

        params: List[Union[np.ndarray, PhiTensor, GammaTensor]] = []
        grads: List[Union[np.ndarray, PhiTensor, GammaTensor]] = []
        for layer in layers:
            params += layer.params
            grads += layer.grads

        if self.ms is None:
            self.ms = [np.zeros(p.shape) for p in params]
        if self.vs is None:
            self.vs = [np.zeros(p.shape) for p in params]

        idx = 0
        for layer in layers:
            new_params: List[Layer] = []

            for p, g in zip(layer.params, layer.grads):
                m = self.ms[idx]
                v = self.vs[idx]
                m = g * (1 - self.beta1) + m * self.beta1
                v = dp_maximum(x=g.abs(), y=(v * self.beta2))  # type: ignore
                p = (m * (-1.0 / (v + self.epsilon)) * a_t) + p
                new_params.append(p)

                self.ms[idx] = m
                self.vs[idx] = v
                idx += 1
            if new_params:
                layer.params = new_params  # type: ignore



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/numpy/tensor.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1cproto/lib/numpy/tensor.proto\x12\x0esyft.lib.numpy";\n\x0cTensorProtos\x12+\n\x06protos\x18\x01 \x03(\x0b\x32\x1b.syft.lib.numpy.TensorProto"\xef\x02\n\x0bTensorProto\x12\x0c\n\x04\x64ims\x18\x01 \x03(\x03\x12\x37\n\tdata_type\x18\x02 \x01(\x0e\x32$.syft.lib.numpy.TensorProto.DataType\x12\x12\n\nfloat_data\x18\x03 \x03(\x01\x12\x12\n\nint32_data\x18\x04 \x03(\x05\x12\x11\n\tbyte_data\x18\x05 \x01(\x0c\x12\x13\n\x0bstring_data\x18\x06 \x03(\x0c\x12\x13\n\x0b\x64ouble_data\x18\x07 \x03(\x01\x12\x12\n\nint64_data\x18\x08 \x03(\x03"\x9f\x01\n\x08\x44\x61taType\x12\r\n\tUNDEFINED\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\t\n\x05INT32\x10\x02\x12\x08\n\x04\x42YTE\x10\x03\x12\n\n\x06STRING\x10\x04\x12\x08\n\x04\x42OOL\x10\x05\x12\t\n\x05UINT8\x10\x06\x12\x08\n\x04INT8\x10\x07\x12\n\n\x06UINT16\x10\x08\x12\t\n\x05INT16\x10\t\x12\t\n\x05INT64\x10\n\x12\x0b\n\x07\x46LOAT16\x10\x0c\x12\n\n\x06\x44OUBLE\x10\rb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.lib.numpy.tensor_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _TENSORPROTOS._serialized_start = 48
    _TENSORPROTOS._serialized_end = 107
    _TENSORPROTO._serialized_start = 110
    _TENSORPROTO._serialized_end = 477
    _TENSORPROTO_DATATYPE._serialized_start = 318
    _TENSORPROTO_DATATYPE._serialized_end = 477
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/request_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.core.node.domain.service import (
    request_message_pb2 as proto_dot_core_dot_node_dot_domain_dot_service_dot_request__message__pb2,
)
from syft.proto.lib.python import dict_pb2 as proto_dot_lib_dot_python_dot_dict__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n*proto/grid/messages/request_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto\x1a\x1bproto/lib/python/dict.proto\x1a\x34proto/core/node/domain/service/request_message.proto"\x9f\x01\n\x14\x43reateRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8c\x01\n\x15\x43reateRequestResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8b\x01\n\x1a\x43reateBudgetRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0e\n\x06\x62udget\x18\x03 \x01(\x02\x12\x0e\n\x06reason\x18\x04 \x01(\t"\xa6\x01\n\x18GetBudgetRequestsMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\nrequest_id\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x92\x01\n\x19GetBudgetRequestsResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x63ontent\x18\x03 \x03(\x0b\x32\x15.syft.lib.python.Dict\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9f\x01\n\x11GetRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\nrequest_id\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8c\x01\n\x12GetRequestResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x12\n\nrequest_id\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8c\x01\n\x12GetRequestsMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xa1\x01\n\x13GetRequestsResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12&\n\x07\x63ontent\x18\x03 \x03(\x0b\x32\x15.syft.lib.python.Dict\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xa2\x01\n\x14\x44\x65leteRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\nrequest_id\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8f\x01\n\x15\x44\x65leteRequestResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x12\n\nrequest_id\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xb2\x01\n\x14UpdateRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0e\n\x06status\x18\x03 \x01(\t\x12\x12\n\nrequest_id\x18\x04 \x01(\t\x12\'\n\x08reply_to\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Address"\x9f\x01\n\x15UpdateRequestResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0e\n\x06status\x18\x03 \x01(\t\x12\x12\n\nrequest_id\x18\x04 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Address"\x8f\x01\n\x15GetAllRequestsMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\xaf\x01\n\x1dGetAllRequestsResponseMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12?\n\x08requests\x18\x03 \x03(\x0b\x32-.syft.core.node.domain.service.RequestMessageb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.grid.messages.request_messages_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CREATEREQUESTMESSAGE._serialized_start = 218
    _CREATEREQUESTMESSAGE._serialized_end = 377
    _CREATEREQUESTRESPONSE._serialized_start = 380
    _CREATEREQUESTRESPONSE._serialized_end = 520
    _CREATEBUDGETREQUESTMESSAGE._serialized_start = 523
    _CREATEBUDGETREQUESTMESSAGE._serialized_end = 662
    _GETBUDGETREQUESTSMESSAGE._serialized_start = 665
    _GETBUDGETREQUESTSMESSAGE._serialized_end = 831
    _GETBUDGETREQUESTSRESPONSE._serialized_start = 834
    _GETBUDGETREQUESTSRESPONSE._serialized_end = 980
    _GETREQUESTMESSAGE._serialized_start = 983
    _GETREQUESTMESSAGE._serialized_end = 1142
    _GETREQUESTRESPONSE._serialized_start = 1145
    _GETREQUESTRESPONSE._serialized_end = 1285
    _GETREQUESTSMESSAGE._serialized_start = 1288
    _GETREQUESTSMESSAGE._serialized_end = 1428
    _GETREQUESTSRESPONSE._serialized_start = 1431
    _GETREQUESTSRESPONSE._serialized_end = 1592
    _DELETEREQUESTMESSAGE._serialized_start = 1595
    _DELETEREQUESTMESSAGE._serialized_end = 1757
    _DELETEREQUESTRESPONSE._serialized_start = 1760
    _DELETEREQUESTRESPONSE._serialized_end = 1903
    _UPDATEREQUESTMESSAGE._serialized_start = 1906
    _UPDATEREQUESTMESSAGE._serialized_end = 2084
    _UPDATEREQUESTRESPONSE._serialized_start = 2087
    _UPDATEREQUESTRESPONSE._serialized_end = 2246
    _GETALLREQUESTSMESSAGE._serialized_start = 2249
    _GETALLREQUESTSMESSAGE._serialized_end = 2392
    _GETALLREQUESTSRESPONSEMESSAGE._serialized_start = 2395
    _GETALLREQUESTSRESPONSEMESSAGE._serialized_end = 2570
# @@protoc_insertion_point(module_scope)



# stdlib
import io
import json
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
import requests
from requests.adapters import HTTPAdapter

# from requests.adapters import TimeoutHTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from requests_toolbelt.multipart.encoder import MultipartEncoder

# relative
from .. import GridURL
from ...core.common.message import ImmediateSyftMessageWithoutReply
from ...core.common.message import SignedImmediateSyftMessageWithoutReply
from ...core.common.message import SyftMessage
from ...core.common.serde.serializable import serializable
from ...core.common.serde.serialize import _serialize
from ...core.node.enums import RequestAPIFields
from ...core.node.exceptions import RequestAPIException
from ...logger import debug
from ...proto.core.node.common.metadata_pb2 import Metadata as Metadata_PB
from ...proto.grid.connections.http_connection_pb2 import (
    GridHTTPConnection as GridHTTPConnection_PB,
)
from ...util import verify_tls
from ..connections.http_connection import HTTPConnection

DEFAULT_TIMEOUT = 30  # seconds


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.timeout = DEFAULT_TIMEOUT
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        if "max_retries" in kwargs:
            self.max_retries = kwargs["max_retries"]
            del kwargs["max_retries"]
        super().__init__(*args, **kwargs)

    def send(self, request: Any, **kwargs: Any) -> Any:  # type:ignore
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


@serializable()
class GridHTTPConnection(HTTPConnection):

    LOGIN_ROUTE = "/login"
    SYFT_ROUTE = "/syft"
    SYFT_ROUTE_STREAM = "/syft/stream"  # non blocking node
    # SYFT_MULTIPART_ROUTE = "/pysyft_multipart"
    SIZE_THRESHOLD = 20971520  # 20 MB

    def __init__(self, url: Union[GridURL, str]) -> None:
        self.base_url = GridURL.from_url(url) if isinstance(url, str) else url
        if self.base_url is None:
            raise Exception(f"Invalid GridURL. {self.base_url}")
        self.session_token: str = ""
        self.token_type: str = "'"

    @property
    def header(self) -> Dict[str, str]:

        _header = {}

        if self.session_token and self.token_type:
            _header = dict(
                Authorization="Bearer "
                + json.loads(
                    '{"auth_token":"'
                    + str(self.session_token)
                    + '","token_type":"'
                    + str(self.token_type)
                    + '"}'
                )["auth_token"]
            )

        _header["Content-Type"] = "application/octet-stream"
        return _header

    def _send_msg(
        self, msg: SyftMessage, timeout: Optional[float] = None
    ) -> requests.Response:
        """
        Serializes Syft messages in json format and send it using HTTP protocol.
        NOTE: Auxiliary method to avoid code duplication and modularity.
        :return: returns requests.Response object containing a JSON serialized
        SyftMessage
        :rtype: requests.Response
        """

        header = self.header

        route = GridHTTPConnection.SYFT_ROUTE
        # if the message has no reply lets use the streaming endpoint
        # this allows the streaming endpoint to run on an entirely different process
        if isinstance(
            msg,
            (SignedImmediateSyftMessageWithoutReply, ImmediateSyftMessageWithoutReply),
        ):
            route = GridHTTPConnection.SYFT_ROUTE_STREAM

        # Perform HTTP request using base_url as a root address
        msg_bytes: bytes = _serialize(obj=msg, to_bytes=True)  # type: ignore

        # timeout = None will wait forever
        timeout = timeout if timeout is not None else DEFAULT_TIMEOUT

        # if sys.getsizeof(msg_bytes) < GridHTTPConnection.SIZE_THRESHOLD:
        # if True:
        r = requests.post(
            url=str(self.base_url) + route,
            data=msg_bytes,
            headers=header,
            verify=verify_tls(),
            timeout=timeout,
            proxies=HTTPConnection.proxies,
        )
        # else:
        #     r = self.send_streamed_messages(blob_message=msg_bytes)

        # Return request's response object
        # r.text provides the response body as a str
        return r

    def login(self, credentials: Dict) -> Tuple:
        response = requests.post(
            url=str(self.base_url) + GridHTTPConnection.LOGIN_ROUTE,
            json=credentials,
            verify=verify_tls(),
            timeout=2,
            proxies=HTTPConnection.proxies,
        )

        # Response
        content = json.loads(response.text)
        # If fail
        if response.status_code != requests.codes.ok:
            raise Exception(content["detail"])

        metadata = content["metadata"].encode("ISO-8859-1")
        metadata_pb = Metadata_PB()
        metadata_pb.ParseFromString(metadata)

        # If success
        # Save session token
        self.session_token = content["access_token"]
        self.token_type = content["token_type"]

        # Return node metadata / user private key
        return (metadata_pb, content["key"])

    def _get_metadata(self, timeout: Optional[float] = 2) -> Tuple:
        """Request Node's metadata
        :return: returns node metadata
        :rtype: str of bytes
        """
        # allow retry when connecting in CI
        session = requests.Session()
        retry = Retry(connect=1, backoff_factor=0.5)
        if timeout is None:
            adapter = HTTPAdapter(max_retries=retry)
        else:
            adapter = TimeoutHTTPAdapter(max_retries=retry, timeout=timeout)
        session.mount("http://", adapter)
        session.mount("https://", adapter)

        metadata_url = str(self.base_url) + "/syft/metadata"
        response = session.get(metadata_url, verify=verify_tls())

        if response.status_code != 200:
            raise requests.ConnectionError(
                f"Failed to fetch metadata. Response returned with code {response.status_code}"
            )

        # upgrade to tls if available
        try:
            if response.url.startswith("https://") and self.base_url.protocol == "http":
                # we got redirected to https
                self.base_url = GridURL.from_url(
                    response.url.replace("/syft/metadata", "")
                )
                debug(f"GridURL Upgraded to HTTPS. {self.base_url}")
        except Exception as e:
            print(f"Failed to upgrade to HTTPS. {e}")

        metadata_pb = Metadata_PB()
        metadata_pb.ParseFromString(response.content)

        return metadata_pb

    def setup(self, **content: Dict[str, Any]) -> Any:
        response = json.loads(
            requests.post(
                str(self.base_url) + "/setup",
                json=content,
                verify=verify_tls(),
                timeout=2,
                proxies=HTTPConnection.proxies,
            ).text
        )
        if response.get(RequestAPIFields.MESSAGE, None):
            return response
        else:
            raise RequestAPIException(response.get(RequestAPIFields.ERROR))

    def reset(self) -> Any:
        header = {}

        if self.session_token and self.token_type:
            header = dict(
                Authorization="Bearer "
                + json.loads(
                    '{"auth_token":"'
                    + str(self.session_token)
                    + '","token_type":"'
                    + str(self.token_type)
                    + '"}'
                )["auth_token"]
            )

        response = json.loads(
            requests.delete(
                str(self.base_url) + GridHTTPConnection.SYFT_ROUTE,
                headers=header,
                verify=verify_tls(),
                proxies=HTTPConnection.proxies,
            ).text
        )
        if response.get(RequestAPIFields.MESSAGE, None, timeout=2):
            return response
        else:
            raise RequestAPIException(response.get(RequestAPIFields.ERROR))

    def send_files(
        self, route: str, file_path: str, form_name: str, form_values: Dict
    ) -> Dict[str, Any]:
        header = {}

        if self.session_token and self.token_type:
            header = dict(
                Authorization="Bearer "
                + json.loads(
                    '{"auth_token":"'
                    + str(self.session_token)
                    + '","token_type":"'
                    + str(self.token_type)
                    + '"}'
                )["auth_token"]
            )

        files = {
            form_name: (None, json.dumps(form_values), "text/plain"),
            "file": (file_path, open(file_path, "rb"), "application/octet-stream"),
        }

        resp = requests.post(
            str(self.base_url) + route,
            files=files,  # type: ignore
            headers=header,
            verify=verify_tls(),
            proxies=HTTPConnection.proxies,
        )

        return json.loads(resp.content)

    def send_streamed_messages(self, blob_message: bytes) -> requests.Response:
        session = requests.Session()
        with io.BytesIO(blob_message) as msg:
            form = MultipartEncoder(
                {
                    "file": ("message", msg.read(), "application/octet-stream"),
                }
            )

            headers = {
                "Prefer": "respond-async",
                "Content-Type": form.content_type,
            }

            resp = session.post(
                str(self.base_url) + GridHTTPConnection.SYFT_ROUTE_STREAM,
                headers=headers,
                data=form,
                verify=verify_tls(),
                proxies=HTTPConnection.proxies,
            )

        session.close()
        return resp

    @property
    def host(self) -> str:
        return self.base_url.base_url

    @staticmethod
    def _proto2object(proto: GridHTTPConnection_PB) -> "GridHTTPConnection":
        obj = GridHTTPConnection(url=GridURL.from_url(proto.base_url))
        obj.session_token = proto.session_token
        obj.token_type = proto.token_type
        return obj

    def _object2proto(self) -> GridHTTPConnection_PB:
        return GridHTTPConnection_PB(
            base_url=str(self.base_url),
            session_token=self.session_token,
            token_type=self.token_type,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return GridHTTPConnection_PB



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/action/get_set_property.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.core.pointer import (
    pointer_pb2 as proto_dot_core_dot_pointer_dot_pointer__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n4proto/core/node/common/action/get_set_property.proto\x12\x1csyft.core.node.common.action\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto\x1a proto/core/pointer/pointer.proto"\xcb\x03\n\x16GetOrSetPropertyAction\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\r\n\x05_self\x18\x02 \x01(\x0c\x12(\n\x04\x61rgs\x18\x03 \x03(\x0b\x32\x1a.syft.core.pointer.Pointer\x12P\n\x06kwargs\x18\x04 \x03(\x0b\x32@.syft.core.node.common.action.GetOrSetPropertyAction.KwargsEntry\x12-\n\x0eid_at_location\x18\x05 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x06 \x01(\x0b\x32\x15.syft.core.io.Address\x12+\n\x07set_arg\x18\x07 \x01(\x0b\x32\x1a.syft.core.pointer.Pointer\x12%\n\x06msg_id\x18\x08 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x0e\n\x06\x61\x63tion\x18\t \x01(\x05\x12\x12\n\nmap_to_dyn\x18\n \x01(\x08\x1aI\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12)\n\x05value\x18\x02 \x01(\x0b\x32\x1a.syft.core.pointer.Pointer:\x02\x38\x01\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.core.node.common.action.get_set_property_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _GETORSETPROPERTYACTION_KWARGSENTRY._options = None
    _GETORSETPROPERTYACTION_KWARGSENTRY._serialized_options = b"8\001"
    _GETORSETPROPERTYACTION._serialized_start = 189
    _GETORSETPROPERTYACTION._serialized_end = 648
    _GETORSETPROPERTYACTION_KWARGSENTRY._serialized_start = 575
    _GETORSETPROPERTYACTION_KWARGSENTRY._serialized_end = 648
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Dict
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft absolute
import syft as sy

# relative
from ..... import lib
from .....proto.core.node.common.action.get_enum_attribute_pb2 import (
    GetEnumAttributeAction as GetEnumAttributeAction_PB,
)
from ....common.serde.serializable import serializable
from ....common.uid import UID
from ....io.address import Address
from ....store.storeable_object import StorableObject
from ...abstract.node import AbstractNode
from .common import ImmediateActionWithoutReply
from .run_class_method_action import RunClassMethodAction


@serializable()
class EnumAttributeAction(ImmediateActionWithoutReply):
    def __init__(
        self,
        path: str,
        id_at_location: UID,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address, msg_id=msg_id)
        self.id_at_location = id_at_location
        self.path = path

    def intersect_keys(
        self,
        left: Dict[VerifyKey, Optional[UID]],
        right: Dict[VerifyKey, Optional[UID]],
    ) -> Dict[VerifyKey, Optional[UID]]:
        return RunClassMethodAction.intersect_keys(left, right)

    def execute_action(self, node: AbstractNode, verify_key: VerifyKey) -> None:
        enum_attribute = node.lib_ast.query(self.path)
        result = enum_attribute.solve_get_enum_attribute().value
        result = lib.python.primitive_factory.PrimitiveFactory.generate_primitive(
            value=result, id=self.id_at_location
        )

        result = StorableObject(
            id=self.id_at_location,
            data=result,
        )

        node.store[self.id_at_location] = result

    def _object2proto(self) -> GetEnumAttributeAction_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetOrSetPropertyAction_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """

        return GetEnumAttributeAction_PB(
            path=self.path,
            id_at_location=sy.serialize(self.id_at_location),
            address=sy.serialize(self.address),
            msg_id=sy.serialize(self.id),
        )

    @staticmethod
    def _proto2object(
        proto: GetEnumAttributeAction_PB,
    ) -> "EnumAttributeAction":
        """Creates a ObjectWithID from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of GetOrSetPropertyAction
        :rtype: GetOrSetPropertyAction
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """
        return EnumAttributeAction(
            path=proto.path,
            id_at_location=sy.deserialize(blob=proto.id_at_location),
            address=sy.deserialize(blob=proto.address),
            msg_id=sy.deserialize(blob=proto.msg_id),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetEnumAttributeAction_PB



# stdlib
import sys

# checks if we are in a python repl or python -i script.py
is_interactive = bool(getattr(sys, "ps1", sys.flags.interactive))

# checks if we are in jupyter
is_jupyter = True

try:
    get_ipython()  # type: ignore
    # third party
    import notebook
    from packaging import version

    NOTEBOOK_VERSION = version.parse(notebook.__version__.split("+")[0])
    if NOTEBOOK_VERSION < version.parse("6.0.0") and "google.colab" not in sys.modules:
        # google.colab check to fix issue #5315
        raise Exception(
            "Your Jupyter Notebook is too old. Please upgrade to version 6 or higher."
        )
except NameError:
    is_jupyter = False

__all__ = ["is_interactive", "is_jupyter"]



# stdlib
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# relative
from ...autodp.gamma_tensor import GammaTensor
from ...autodp.phi_tensor import PhiTensor


class Layer:
    """
    Subclassed when implementing new types of layers.

    Each layer can keep track of the layer(s) feeding into it, a
    network's output :class:`Layer` instance can double as a handle to the full
    network.
    """

    first_layer: bool = False

    def forward(
        self,
        input: Union[PhiTensor, GammaTensor],
        *args: Optional[Any],
        **kwargs: Optional[Any]
    ) -> Union[PhiTensor, GammaTensor]:
        raise NotImplementedError

    def backward(
        self,
        pre_grad: Union[PhiTensor, GammaTensor],
        *args: Optional[Any],
        **kwargs: Optional[Any]
    ) -> Optional[Union[PhiTensor, GammaTensor]]:
        raise NotImplementedError

    def connect_to(self, prev_layer: "Layer") -> None:
        raise NotImplementedError

    @property
    def params(self) -> Tuple:
        """Layer parameters.

        Returns a list of numpy.array variables or expressions that
        parameterize the layer.
        Returns
        -------
        list of numpy.array variables or expressions
            A list of variables that parameterize the layer
        Notes
        -----
        For layers without any parameters, this will return an empty list.
        """
        return ()

    @property
    def grads(self) -> Tuple:
        """Get layer parameter gradients as calculated from backward()."""
        return ()

    @property
    def param_grads(self) -> List:
        """Layer parameters and corresponding gradients."""
        return list(zip(self.params, self.grads))

    def __str__(self) -> str:
        return self.__class__.__name__



# stdlib
from typing import Any
from typing import List


class VMMessageRegistry:
    """A class for registering messages that can be used by the VirtualMachine service."""

    __vm_message_registry: List = []

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        cls.__vm_message_registry.append(cls)

    @classmethod
    def get_registered_messages(cls) -> List[Any]:
        """dict: Returns map of service inheriting this class."""
        return list(cls.__vm_message_registry)



# future
from __future__ import annotations

# stdlib
import json
import os
import re
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
from nacl.signing import VerifyKey
import requests
from typing_extensions import final

# relative
from ......grid import GridURL
from ......logger import critical
from ......util import verify_tls
from .....common.serde.serializable import serializable
from ....abstract.node import AbstractNode
from ..generic_payload.messages import GenericPayloadMessage
from ..generic_payload.messages import GenericPayloadMessageWithReply
from ..generic_payload.messages import GenericPayloadReplyMessage

TAILSCALE_HOST = "proxy"
TAILSCALE_PORT = 4000
TAILSCALE_URL = f"http://{TAILSCALE_HOST}:{TAILSCALE_PORT}"

HEADSCALE_HOST = "headscale"
HEADSCALE_PORT = 4000
HEADSCALE_URL = f"http://{HEADSCALE_HOST}:{HEADSCALE_PORT}"


def grid_url_from_kwargs(kwargs: Dict[str, Any]) -> GridURL:
    try:
        if "host_or_ip" in kwargs:
            # old way to send these messages was with host_or_ip
            return GridURL.from_url(str(kwargs["host_or_ip"]))
        elif "grid_url" in kwargs:
            # new way is with grid_url
            return kwargs["grid_url"]
        else:
            raise Exception("kwargs missing host_or_ip or grid_url")
    except Exception as e:
        print(f"Failed to get grid_url from kwargs: {kwargs}. {e}")
        raise e


@serializable(recursive_serde=True)
@final
class VPNConnectMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNConnectReplyMessage(GenericPayloadReplyMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNConnectMessageWithReply(GenericPayloadMessageWithReply):
    message_type = VPNConnectMessage
    message_reply_type = VPNConnectReplyMessage

    def run(
        self, node: AbstractNode, verify_key: Optional[VerifyKey] = None
    ) -> Dict[str, Any]:
        try:
            grid_url = grid_url_from_kwargs(self.kwargs)
            grid_url = grid_url.with_path("/vpn")
            vpn_auth_key = str(self.kwargs["vpn_auth_key"])

            status, error = disconnect(tailscale_host=TAILSCALE_URL)
            if not status:
                print("Failed to run tailscale down first", error)

            status, error = connect_with_key(
                tailscale_host=TAILSCALE_URL,
                headscale_host=str(grid_url),
                vpn_auth_key=vpn_auth_key,
            )
            if status:
                return {"status": "ok"}
            else:
                print("connect with key failed", error)
                return {"status": "error"}
        except Exception as e:
            print(f"Failed to run {type(self)}", self.kwargs, e)
            return {"status": "error"}


def disconnect(tailscale_host: str) -> Tuple[bool, str]:
    try:
        # we need --accept-dns=false because magicDNS replaces /etc/resolv.conf which
        # breaks using tailscale in network_mode with docker compose because the
        # /etc/resolv.conf has the mDNS ip nameserver 127.0.0.11
        data = {
            "args": [],
            "timeout": 60,
            "force_unique_key": True,
        }

        command_url = f"{tailscale_host}/commands/down"

        headers = {"X-STACK-API-KEY": os.environ.get("STACK_API_KEY", "")}
        resp = requests.post(command_url, json=data, headers=headers)
        report = get_result(json=resp.json())
        report_dict = json.loads(report)

        if int(report_dict["returncode"]) == 0:
            return (True, "")
        else:
            return (False, report_dict.get("report", ""))
    except Exception as e:
        critical(f"Failed to disconnect VPN. {e}")
        raise e


def connect_with_key(
    tailscale_host: str, headscale_host: str, vpn_auth_key: str
) -> Tuple[bool, str]:
    try:
        # we need --accept-dns=false because magicDNS replaces /etc/resolv.conf which
        # breaks using tailscale in network_mode with docker compose because the
        # /etc/resolv.conf has the mDNS ip nameserver 127.0.0.11
        data = {
            "args": [
                "-login-server",
                f"{headscale_host}",
                "--reset",
                "--force-reauth",
                "--authkey",
                f"{vpn_auth_key}",
                "--accept-dns=false",
            ],
            "timeout": 60,
            "force_unique_key": True,
        }

        command_url = f"{tailscale_host}/commands/up"

        headers = {"X-STACK-API-KEY": os.environ.get("STACK_API_KEY", "")}
        resp = requests.post(command_url, json=data, headers=headers)
        report = get_result(json=resp.json())
        report_dict = json.loads(report)

        if int(report_dict["returncode"]) == 0:
            return (True, "")
        else:
            return (False, report_dict.get("report", ""))
    except Exception as e:
        critical(f"Failed to connect to VPN. {e}")
        raise e


@serializable(recursive_serde=True)
@final
class VPNJoinMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNJoinReplyMessage(GenericPayloadReplyMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNJoinMessageWithReply(GenericPayloadMessageWithReply):
    message_type = VPNJoinMessage
    message_reply_type = VPNJoinReplyMessage

    def run(
        self, node: AbstractNode, verify_key: Optional[VerifyKey] = None
    ) -> Dict[str, Any]:
        try:
            # can't import Network due to circular imports
            if type(node).__name__ == "Network":
                # networks can't join other networks yet
                raise Exception(
                    "Network cant join another Network, try VPNJoinSelfMessageWithReply"
                )

            # we are running inside the container so we should change the host to
            # what ever will suit the environment with as_container_host
            grid_url = grid_url_from_kwargs(self.kwargs).as_container_host(
                container_host=node.settings.CONTAINER_HOST
            )

            res = requests.post(
                str(grid_url.with_path("/api/v1/vpn/register")), verify=verify_tls()
            )
            res_json = res.json()

            if "vpn_auth_key" not in res_json:
                print("Registration failed", res)
                return {"status": "error"}

            status, error = disconnect(tailscale_host=TAILSCALE_URL)
            if not status:
                print("Failed to run tailscale down first", error)

            status, error = connect_with_key(
                tailscale_host=TAILSCALE_URL,
                headscale_host=str(grid_url.with_path("/vpn")),
                vpn_auth_key=res_json["vpn_auth_key"],
            )

            if status:
                return {"status": "ok"}
            else:
                print("connect with key failed", error)
                return {"status": "error"}
        except Exception as e:
            print(f"Failed to run {type(self)}", self.kwargs, e)
            return {"status": "error"}


@serializable(recursive_serde=True)
@final
class VPNJoinSelfMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNJoinSelfReplyMessage(GenericPayloadReplyMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNJoinSelfMessageWithReply(GenericPayloadMessageWithReply):
    message_type = VPNJoinSelfMessage
    message_reply_type = VPNJoinSelfReplyMessage

    def run(
        self, node: AbstractNode, verify_key: Optional[VerifyKey] = None
    ) -> Dict[str, Any]:
        try:
            # can't import Domaiun due to circular imports
            if type(node).__name__ == "Domain":
                # networks can't join other networks yet
                raise Exception(
                    "Domains cant join themselves, try VPNJoinMessageWithReply"
                )

            # get status
            up_status, _, _ = get_status(tailscale_host=TAILSCALE_URL)
            if up_status:
                print("Already connected!")
                return {"status": "ok"}
            print("Connecting...")

            key_status, vpn_auth_key = generate_key(headscale_host=HEADSCALE_URL)

            if not key_status:
                raise Exception("Failed to generate key for joining VPN")

            status, error = disconnect(tailscale_host=TAILSCALE_URL)
            if not status:
                print("Failed to run tailscale down first", error)

            connect_status, error = connect_with_key(
                tailscale_host=TAILSCALE_URL,
                headscale_host=f"http://{HEADSCALE_HOST}:8080",
                vpn_auth_key=vpn_auth_key,
            )

            if connect_status:
                print("Connected successfully")
                return {"status": "ok"}
            else:
                print("connect with key failed", error)
                return {"status": "error"}
        except Exception as e:
            print(f"Failed to run {type(self)}", self.kwargs, e)
            return {"status": "error"}


@serializable(recursive_serde=True)
@final
class VPNRegisterMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNRegisterReplyMessage(GenericPayloadReplyMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNRegisterMessageWithReply(GenericPayloadMessageWithReply):
    message_type = VPNRegisterMessage
    message_reply_type = VPNRegisterReplyMessage

    def run(self, node: AbstractNode, verify_key: Optional[VerifyKey] = None) -> Any:
        try:
            status, vpn_auth_key = generate_key(headscale_host=HEADSCALE_URL)

            if status:
                return {"status": "ok", "vpn_auth_key": vpn_auth_key}
            else:
                print("register and create vpn_auth_key failed")
                return {"status": "error"}
        except Exception as e:
            print(f"Failed to run {type(self)}", self.kwargs, e)
            return {"status": "error"}


# the response contains a dict with a key called "report" which contains unescaped json
def extract_nested_json(nested_json: str) -> Union[Dict, List]:
    matcher = r'"report":"(.+)\\n"'
    parts = re.findall(matcher, nested_json)
    report_json = parts[0].replace("\\n", "").replace("\\t", "").replace('\\"', '"')
    return json.loads(report_json)


def generate_key(headscale_host: str) -> Tuple[bool, str]:
    data = {"timeout": 5, "force_unique_key": True}

    command_url = f"{headscale_host}/commands/generate_key"
    try:
        headers = {"X-STACK-API-KEY": os.environ.get("STACK_API_KEY", "")}
        resp = requests.post(command_url, json=data, headers=headers)
        report = get_result(json=resp.json())
        result_dict = dict(extract_nested_json(report))
        result = result_dict["key"]

        # check if we got a key
        if len(result) == 48:
            return (True, result)
        else:
            return (False, "")

    except Exception as e:
        print("failed to make request", e)
        return (False, str(e))


@serializable(recursive_serde=True)
@final
class VPNStatusMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNStatusReplyMessage(GenericPayloadReplyMessage):
    ...


@serializable(recursive_serde=True)
@final
class VPNStatusMessageWithReply(GenericPayloadMessageWithReply):
    message_type = VPNStatusMessage
    message_reply_type = VPNStatusReplyMessage

    def run(self, node: AbstractNode, verify_key: Optional[VerifyKey] = None) -> Any:
        try:
            up, host, peers = get_status(tailscale_host=TAILSCALE_URL)
            return {"status": "ok", "connected": up, "host": host, "peers": peers}
        except Exception as e:
            print(f"Failed to run {type(self)}", self.kwargs, e)
            return {"status": "error"}


def clean_status_output(
    input: str,
) -> Tuple[bool, Dict[str, str], List[Dict[str, str]]]:
    # example input
    """
    # Health check:
    #     - dns: rename /etc/resolv.conf /etc/resolv.pre-tailscale-backup.conf: device or resource busy

    100.64.0.1      test_domain_1        omnet        linux   -
    100.64.0.2      test_network_1       omnet        linux   active; relay "syd", tx 1188 rx 1040
    """
    up = False
    peers: List[Dict[str, str]] = []
    host: Dict[str, str] = {}
    if "Tailscale is stopped." in input:
        return up, host, peers
    elif "unexpected state: NoState" in input:
        return up, host, peers

    count = 0
    for line in str(input).split("\n"):
        matches = re.match(r"^\d.+", line)
        if matches is not None:
            try:
                stat_parts = re.split(r"(\s+)", matches.string)
                entry = {}
                entry["ip"] = stat_parts[0]
                entry["hostname"] = stat_parts[2]
                entry["network"] = stat_parts[4]
                entry["os"] = stat_parts[6]
                connection_info_parts = matches.string.split(entry["os"])
                entry["connection_info"] = "n/a"
                connection_info = ""
                if len(connection_info_parts) > 1:
                    connection_info = connection_info_parts[1].strip()
                    entry["connection_info"] = connection_info

                entry["connection_status"] = "n/a"
                if "active" in connection_info:
                    entry["connection_status"] = "active"

                if "idle" in connection_info:
                    entry["connection_status"] = "idle"

                entry["connection_type"] = "n/a"
                if "relay" in connection_info:
                    entry["connection_type"] = "relay"

                if "direct" in connection_info:
                    entry["connection_type"] = "direct"

                if count == 0:
                    host = entry
                    count += 1
                    up = True
                else:
                    peers.append(entry)

            except Exception as e:
                print("Error parsing tailscale status output", e)
                pass

    return up, host, peers


def get_status(
    tailscale_host: str,
) -> Tuple[bool, Dict[str, str], List[Dict[str, str]]]:
    data = {"timeout": 5, "force_unique_key": True}
    command_url = f"{tailscale_host}/commands/status"
    host: Dict[str, str] = {}
    peers: List[Dict[str, str]] = []
    connected = False
    try:
        headers = {"X-STACK-API-KEY": os.environ.get("STACK_API_KEY", "")}
        resp = requests.post(command_url, json=data, headers=headers)
        report = get_result(json=resp.json())
        cmd_output = json.loads(report)["report"]
        connected, host, peers = clean_status_output(input=cmd_output)
    except Exception as e:
        print("failed to make request", e)
    return connected, host, peers


def get_result(json: Dict) -> str:
    result_url = json.get("result_url", "")
    headers = {"X-STACK-API-KEY": os.environ.get("STACK_API_KEY", "")}
    tries = 0
    limit = 5
    try:
        while True:
            print("Polling API Result", tries)
            result = requests.get(result_url, headers=headers)
            if '"status":"running"' in result.text:
                time.sleep(1)
                tries += 1
                if tries > limit:
                    break
                continue
            else:
                return str(result.text)
    except Exception as e:
        print("Failed to get result", json, e)
    return "{}"



# third party
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String

# relative
from . import Base


class Role(Base):
    __tablename__ = "role"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    name = Column(String(255))
    can_make_data_requests = Column(Boolean(), default=False)
    can_triage_data_requests = Column(Boolean(), default=False)
    can_manage_privacy_budget = Column(Boolean(), default=False)
    can_create_users = Column(Boolean(), default=False)
    can_manage_users = Column(Boolean(), default=False)
    can_edit_roles = Column(Boolean(), default=False)
    can_manage_infrastructure = Column(Boolean(), default=False)
    can_upload_data = Column(Boolean(), default=False)
    can_upload_legal_document = Column(Boolean(), default=False)
    can_edit_domain_settings = Column(Boolean(), default=False)

    def __str__(self) -> str:
        return (
            f"<Role id: {self.id}, name: {self.name}, "
            f"can_make_data_requests: {self.can_make_data_requests}, "
            f"can_triage_data_requests: {self.can_triage_data_requests}, "
            f"can_manage_privacy_budget: {self.can_manage_privacy_budget}, "
            f"can_create_users: {self.can_create_users}, "
            f"can_manage_users: {self.can_manage_users}, "
            f"can_edit_roles: {self.can_edit_roles}>, "
            f"can_manage_infrastructure: {self.can_manage_infrastructure}>"
            f"can_upload_data: {self.can_upload_data}>"
        )


def create_role(
    name: str,
    can_make_data_requests: bool,
    can_triage_data_requests: bool,
    can_manage_privacy_budget: bool,
    can_create_users: bool,
    can_manage_users: bool,
    can_edit_roles: bool,
    can_manage_infrastructure: bool,
    can_upload_data: bool,
    can_upload_legal_document: bool,
    can_edit_domain_settings: bool,
) -> Role:
    new_role = Role(
        name=name,  # type:ignore
        can_make_data_requests=can_make_data_requests,
        can_triage_data_requests=can_triage_data_requests,
        can_manage_privacy_budget=can_manage_privacy_budget,
        can_create_users=can_create_users,
        can_manage_users=can_manage_users,
        can_edit_roles=can_edit_roles,
        can_manage_infrastructure=can_manage_infrastructure,
        can_upload_data=can_upload_data,
        can_upload_legal_document=can_upload_legal_document,
        can_edit_domain_settings=can_edit_domain_settings,
    )
    return new_role



# stdlib
from datetime import datetime
from typing import Any
from typing import Dict
from typing import Optional

# third party
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker

# relative
from ....common.uid import UID
from ..exceptions import RequestError
from ..node_service.request_receiver.request_receiver_messages import RequestStatus
from ..node_table.request import Request
from .database_manager import DatabaseManager


class RequestManager(DatabaseManager):

    schema = Request

    def __init__(self, database: Engine) -> None:
        super().__init__(schema=RequestManager.schema, db=database)

    def first(self, **kwargs: Any) -> Request:
        result = super().first(**kwargs)
        if not result:
            raise RequestError(
                "Can't process AcceptOrDenyRequestService using this request ID."
            )

        return result

    def create_request(self, **kwargs: Any) -> Request:
        date = datetime.now()
        return self.register(id=str(UID().value), date=date, **kwargs)

    def status(self, request_id: str) -> RequestStatus:
        _req = self.first(id=request_id)
        if _req.status == "pending":
            return RequestStatus.Pending
        elif _req.status == "accepted":
            return RequestStatus.Accepted
        else:
            return RequestStatus.Rejected

    def set(self, request_id: int, status: RequestStatus) -> None:
        self.modify({"id": request_id}, {"status": status})

    def get_user_info(self, request_id: int) -> Dict:
        request: Optional[Request] = super().first(id=request_id)
        if request:
            return {
                "name": request.user_name,
                "email": request.user_email,
                "role": request.user_role,
                "current_budget": request.user_budget,
                "institution": request.institution,
                "website": request.website,
            }
        else:
            return {}

    def get_req_info(self, request_id: int) -> Dict:
        request: Optional[Request] = super().first(id=request_id)
        if request:
            return {
                "id": request.id,
                "date": str(request.date),
                "status": request.status,
                "reason": request.reason,
                "request_type": request.request_type,
                "current_budget": request.current_budget,
                "requested_budget": request.requested_budget,
                "review": {
                    "name": request.reviewer_name,
                    "role": request.reviewer_role,
                    # TODO: fix datetime
                    # "updated_on": str(request.updated_on),
                    "comment": request.reviewer_comment,
                },
            }
        else:
            return {}

    def clear(self) -> None:
        local_session = sessionmaker(bind=self.db)()
        local_session.query(self.schema).delete()
        local_session.commit()
        local_session.close()



# third party
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String

# relative
from . import Base


class BinObjDataset(Base):
    __tablename__ = "bin_obj_dataset"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    name = Column(String(256))
    obj = Column(String(256))
    dataset = Column(String(256), ForeignKey("dataset.id", ondelete="CASCADE"))
    dtype = Column(String(256))
    shape = Column(String(256))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/python/iterator.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1fproto/lib/python/iterator.proto\x12\x0fsyft.lib.python\x1a%proto/core/common/common_object.proto"q\n\x08Iterator\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x0f\n\x07obj_ref\x18\x02 \x01(\x0c\x12\r\n\x05index\x18\x03 \x01(\x05\x12\x0f\n\x07max_len\x18\x04 \x01(\x05\x12\x11\n\texhausted\x18\x05 \x01(\x08\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.lib.python.iterator_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _ITERATOR._serialized_start = 91
    _ITERATOR._serialized_end = 204
# @@protoc_insertion_point(module_scope)



"""function used to calculate reciprocal of a given tensor."""

# future
from __future__ import annotations

# stdlib
from typing import TYPE_CHECKING

# relative
from .exp import exp
from .log import log
from .utils import modulus
from .utils import sign

if TYPE_CHECKING:
    # relative
    from ...tensor.smpc.mpc_tensor import MPCTensor


def reciprocal(data: MPCTensor, method: str = "NR", nr_iters: int = 5) -> MPCTensor:
    """Calculate the reciprocal using the algorithm specified in the method args.
    Ref: https://github.com/facebookresearch/CrypTen

    Args:
        data: input data
        nr_iters: Number of iterations for Newton-Raphson
        method: 'NR' - `Newton-Raphson`_ method computes the reciprocal using iterations
                of :math:`x_{i+1} = (2x_i - data * x_i^2)` and uses
                :math:`3*exp(-(x-.5)) + 0.003` as an initial guess by default
                'log' -  Computes the reciprocal of the input from the observation that:
                        :math:`x^{-1} = exp(-log(x))`

    Returns:
        Reciprocal of `data`

    Raises:
        ValueError: if the given method is not supported
    """
    method = method.lower()

    if method == "nr":
        new_data = modulus(data)
        result = exp(new_data * -1 + 0.5) * 3 + 0.003
        for i in range(nr_iters):
            result = result * 2 - result * result * new_data
        return result * sign(data)
    elif method == "log":
        new_data = modulus(data)
        return exp(-1 * log(new_data)) * sign(data)
    else:
        raise ValueError(f"Invalid method {method} given for reciprocal function")



# stdlib
from collections import UserDict
from collections.abc import ItemsView
from collections.abc import KeysView
from collections.abc import ValuesView
from typing import Any
from typing import Dict as TypeDict
from typing import Iterable
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# relative
from ...core.common import UID
from ...core.common.serde.serializable import serializable
from ...logger import traceback_and_raise
from ...logger import warning
from ...proto.lib.python.dict_pb2 import Dict as Dict_PB
from .iterator import Iterator
from .primitive_factory import PrimitiveFactory
from .primitive_factory import isprimitive
from .primitive_interface import PyPrimitive
from .types import SyPrimitiveRet
from .util import downcast
from .util import upcast


@serializable()
class Dict(UserDict, PyPrimitive):
    # the incoming types to UserDict __init__ are overloaded and weird
    # see https://github.com/python/cpython/blob/master/Lib/collections/__init__.py
    # this is the version from python 3.7 because we need to support 3.7
    # python 3.8 signature includes a new PEP 570 (args, /, kwargs) syntax:
    # https://www.python.org/dev/peps/pep-0570/

    def __init__(*args: Any, **kwargs: Any) -> None:
        if not args:
            traceback_and_raise(
                TypeError("descriptor '__init__' of 'Dict' object " "needs an argument")
            )
        self, *args = args  # type: ignore
        if len(args) > 1:
            traceback_and_raise(
                TypeError(f"expected at most 1 arguments, got {len(args)}")
            )
        if args:
            args_dict = args[0]
        elif "dict" in kwargs:
            args_dict = kwargs.pop("dict")

            warning(
                "Passing 'dict' as keyword argument is deprecated",
                DeprecationWarning,
                stacklevel=2,
            )
        else:
            args_dict = None

        self.data = {}
        if args_dict is not None:
            self.update(args_dict)
        if kwargs:
            self.update(kwargs)

        # We cant add UID from kwargs or it could easily be overwritten by the dict
        # that is being passed in for __init__
        # If you want to update it use the _id setter after creation.
        self._id = UID()

        temporary_box = kwargs["temporary_box"] if "temporary_box" in kwargs else False
        if temporary_box:
            PyPrimitive.__init__(
                self,
                temporary_box=temporary_box,
            )

    @property
    def id(self) -> UID:
        """We reveal PyPrimitive.id as a property to discourage users and
        developers of Syft from modifying .id attributes after an object
        has been initialized.

        :return: returns the unique id of the object
        :rtype: UID
        """
        return self._id

    def upcast(self) -> TypeDict:
        # recursively upcast
        result = {k: upcast(v) for k, v in self.items()}
        if "temporary_box" in result:
            del result["temporary_box"]
        return result

    def __contains__(self, other: Any) -> SyPrimitiveRet:
        res = super().__contains__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __eq__(self, other: Any) -> SyPrimitiveRet:
        res = super().__eq__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __format__(self, format_spec: str) -> str:
        # python complains if the return value is not str
        res = super().__format__(format_spec)
        return str(res)

    def __ge__(self, other: Any) -> SyPrimitiveRet:
        res = super().__ge__(other)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def __getitem__(self, key: Any) -> Union[SyPrimitiveRet, Any]:
        res = super().__getitem__(key)
        if isprimitive(value=res):
            return PrimitiveFactory.generate_primitive(value=res)
        else:
            # we can have torch.Tensor and other types
            return res

    def __gt__(self, other: Any) -> SyPrimitiveRet:
        res = super().__gt__(other)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def __hash__(self) -> SyPrimitiveRet:
        res = super().__hash__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __iter__(self, max_len: Optional[int] = None) -> Iterator:
        return Iterator(super().__iter__(), max_len=max_len)

    def __le__(self, other: Any) -> SyPrimitiveRet:
        res = super().__le__(other)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def __len__(self) -> SyPrimitiveRet:
        res = super().__len__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __lt__(self, other: Any) -> SyPrimitiveRet:
        res = super().__lt__(other)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def __ne__(self, other: Any) -> SyPrimitiveRet:
        res = super().__ne__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __sizeof__(self) -> SyPrimitiveRet:
        res = super().__sizeof__()
        return PrimitiveFactory.generate_primitive(value=res)

    def copy(self) -> SyPrimitiveRet:
        res = super().copy()
        return PrimitiveFactory.generate_primitive(value=res)

    @classmethod
    def fromkeys(
        cls, iterable: Iterable, value: Optional[Any] = None
    ) -> SyPrimitiveRet:
        res = super().fromkeys(iterable, value)
        return PrimitiveFactory.generate_primitive(value=res)

    def dict_get(self, key: Any, default: Any = None) -> Any:
        res = super().get(key, default)
        if isprimitive(value=res):
            return PrimitiveFactory.generate_primitive(value=res)
        else:
            # we can have torch.Tensor and other types
            return res

    def items(self, max_len: Optional[int] = None) -> Iterator:  # type: ignore
        return Iterator(ItemsView(self), max_len=max_len)

    def keys(self, max_len: Optional[int] = None) -> Iterator:  # type: ignore
        return Iterator(KeysView(self), max_len=max_len)

    def values(self, *args: Any, max_len: Optional[int] = None) -> Iterator:  # type: ignore
        # this is what the super type does and there is a test in dict_test.py
        # test_values which checks for this so we could disable the test or
        # keep this workaround
        if len(args) > 0:
            traceback_and_raise(
                TypeError("values() takes 1 positional argument but 2 were given")
            )
        return Iterator(ValuesView(self), max_len=max_len)

    def pop(self, key: Any, *args: Any) -> SyPrimitiveRet:
        res = super().pop(key, *args)
        return PrimitiveFactory.generate_primitive(value=res)

    def popitem(self) -> SyPrimitiveRet:
        res = self.data.popitem()
        return PrimitiveFactory.generate_primitive(value=res)

    def setdefault(self, key: Any, default: Any = None) -> SyPrimitiveRet:
        res = PrimitiveFactory.generate_primitive(value=default)
        res = super().setdefault(key, res)
        return res

    def clear(self) -> None:
        # we get the None return and create a SyNone
        # this is to make sure someone doesn't rewrite the method to return nothing
        return PrimitiveFactory.generate_primitive(value=super().clear())

    def _object2proto(self) -> Dict_PB:
        id_ = sy.serialize(obj=self.id)

        keys = [
            sy.serialize(obj=downcast(value=element), to_bytes=True)
            for element in self.data.keys()
        ]

        values = [
            sy.serialize(obj=downcast(value=element), to_bytes=True)
            for element in self.data.values()
        ]

        if hasattr(self, "temporary_box"):
            temporary_box = self.temporary_box
        else:
            temporary_box = False

        return Dict_PB(
            id=id_,
            keys=keys,
            values=values,
            temporary_box=temporary_box,
        )

    @staticmethod
    def _proto2object(proto: Dict_PB) -> "Dict":
        id_: UID = sy.deserialize(blob=proto.id)

        values = [
            upcast(value=sy.deserialize(blob=element, from_bytes=True))
            for element in proto.values
        ]

        keys = [
            upcast(value=sy.deserialize(blob=element, from_bytes=True))
            for element in proto.keys
        ]
        new_dict = Dict(dict(zip(keys, values)))
        new_dict.temporary_box = proto.temporary_box
        new_dict._id = id_
        return new_dict

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Dict_PB



# stdlib
from io import BytesIO
from typing import Any
from typing import Generator
from typing import List

# third party
import boto3
from botocore.client import Config
from pydantic import BaseSettings
import requests
from tqdm import tqdm

# relative
from ....grid import GridURL
from ....util import get_fully_qualified_name
from ....util import size_mb
from ...common.serde.serialize import _serialize as serialize
from ...common.uid import UID
from ...store.proxy_dataset import ProxyDataset
from .exceptions import DatasetUploadError

MIN_BLOB_UPLOAD_SIZE_MB = 1


def read_chunks(
    fp: BytesIO, chunk_size: int = 1024**3
) -> Generator[bytes, None, None]:
    """Read data in chunks from the file."""
    while True:
        data = fp.read(chunk_size)
        if not data:
            break
        yield data


def listify(x: Any) -> List[Any]:
    """turns x into a list.
    If x is a list or tuple, return as list.
    if x is not a list: return [x]
    if x is None: return []

    Args:
        x (Any): some object

    Returns:
        List[Any]: x, as a list
    """
    return list(x) if isinstance(x, (list, tuple)) else ([] if x is None else [x])


def upload_result_to_s3(
    asset_name: str,
    dataset_name: str,
    domain_id: UID,
    data: Any,
    settings: BaseSettings,
) -> ProxyDataset:
    """Upload data to Seaweed using boto3 client.

    - Serialize data to binary
    - Upload data to Seaweed using boto3 client
    - Create a ProxyDataset to store the metadata of the data uploaded to Seaweed

    Args:
        asset_name (str): name of the data being uploaded to Seaweed
        dataset_name (str): name of the dataset to which the data belongs
        domain_id (UID): unique id of the domain node
        data (Any): data to be uploaded to Seaweed
        settings (BaseSettings): base settings of the PyGrid server

    Returns:
        ProxyDataset: Class to store the metadata of the data being uploaded
    """

    s3_client = get_s3_client(settings=settings)

    binary_dataset: bytes = serialize(data, to_bytes=True)  # type: ignore

    # 1- Serialize the data to be uploaded to bytes.
    binary_buffer = BytesIO(binary_dataset)
    filename = f"{dataset_name}/{asset_name}"

    # 2 - Start to upload binary data into Seaweed.
    upload_response = s3_client.put_object(
        Bucket=domain_id.no_dash,
        Body=binary_buffer,
        Key=filename,
        ContentType="application/octet-stream",
    )

    # TODO: Throw an exception if the response is not valid
    print("Upload Result")
    print(upload_response)

    # 3 - Create a ProxyDataset for the given data
    # Retrieve fully qualified name to  use for pointer creation.
    obj_public_kwargs = getattr(data, "proxy_public_kwargs", {})
    data_fqn = str(get_fully_qualified_name(data))
    data_dtype = str(type(data))

    # relative
    from ...tensor import Tensor
    from ...tensor.autodp.phi_tensor import PhiTensor as PT

    if isinstance(data, Tensor) and isinstance(data.child, PT):
        data_fqn = str(get_fully_qualified_name(data.child))
        child_kwargs = getattr(data.child, "proxy_public_kwargs", {})
        obj_public_kwargs.update(child_kwargs)

    proxy_obj = ProxyDataset(
        asset_name=asset_name,
        dataset_name=dataset_name,
        node_id=domain_id,
        dtype=data_dtype,
        fqn=data_fqn,
        shape=data.shape,
        obj_public_kwargs=obj_public_kwargs,
    )
    return proxy_obj


def abort_s3_object_upload(
    client: boto3.client, upload_id: str, asset_name: str
) -> None:
    """Abort upload to s3 for the given asset.

    Args:
        client (boto3.client): boto3 client.
        upload_id (str): upload id generated for mutlipart upload.
        asset_name (str): name of the data/asset.
    """
    # relative
    from .node_service.upload_service.upload_service_messages import (
        AbortDataUploadMessage,
    )

    # Send a message to PyGrid to abort the data being uploaded to s3.
    # TODO: make this call async -> could use celery .delay method to do
    _ = client.datasets.perform_api_request_generic(
        syft_msg=AbortDataUploadMessage,
        content={
            "upload_id": upload_id,
            "asset_name": asset_name,
        },
    )


def upload_to_s3_using_presigned(
    client: Any,
    data: Any,
    chunk_size: int,
    asset_name: str,
    dataset_name: str = "",
) -> ProxyDataset:
    """Perform a multipart upload of data to Seaweed using boto3 presigned urls.
    The main steps involve:
    - Converting the data to binary data
    - Chunking the binary into smaller chunks
    - Create presigned urls for each chunk
    - Upload data to Seaweed via PUT request
    - Send a acknowledge to Seaweed via PyGrid when all chunks are successfully uploaded
    - Create a ProxyDataset to store metadata of the uploaded data
    Args:
        client (Any): Client to send object to
        data (Any): Data to be uploaded to Seaweed
        chunk_size (int): smallest size of the data to be uploaded in bytes
        asset_name (str): name of the data being uploaded
        dataset_name Optional[(str)]: name of the dataset to which the data belongs
    Raises:
        Exception: If upload of data chunks to Seaweed fails.
    Returns:
        ProxyDataset: Class to store metadata about the data that is uploaded to Seaweed.
    """
    data_upload_description = f"Uploading `{asset_name}`"
    # relative
    from .node_service.upload_service.upload_service_messages import (
        UploadDataCompleteMessage,
    )
    from .node_service.upload_service.upload_service_messages import UploadDataMessage

    dataset_name = str(dataset_name)

    # Step 1 - Convert data to be uploaded to binary
    binary_dataset: bytes = serialize(data, to_bytes=True)  # type: ignore
    file_size = len(binary_dataset)

    # Step 2 - Send a message to PyGrid to inform of the data being uploaded,
    # and get presigned url for each chunk of data being uploaded.
    upload_response = client.datasets.perform_api_request_generic(
        syft_msg=UploadDataMessage,
        content={
            "filename": f"{dataset_name}/{asset_name}",
            "file_size": file_size,
            "chunk_size": chunk_size,
            "address": client.address,
            "reply_to": client.address,
        },
    )

    try:
        # Step 3 - Starts to upload binary data into Seaweed.
        binary_buffer = BytesIO(binary_dataset)
        parts = sorted(upload_response.payload.parts, key=lambda x: x["part_no"])
        etag_chunk_no_pairs = list()
        data_chunks = zip(read_chunks(binary_buffer, chunk_size), parts)
        for _ in tqdm(
            parts, desc=data_upload_description, colour="green", ncols=100, leave=True
        ):
            data_chunk, part = next(data_chunks)
            presigned_url = part["url"]
            part_no = part["part_no"]
            client_url = client.url_from_path(presigned_url)
            part["client_url"] = client_url

            res = requests.put(client_url, data=data_chunk)

            if not res.ok:  # raise an error if upload fails
                error_message = (
                    f"\n\nFailed to upload `{asset_name}` to store\n"
                    + f"Status code: {res.status_code} {res.reason}\n"
                    + f"Error: {str(res.text)}"
                )
                raise DatasetUploadError(message=error_message)
            etag = res.headers["ETag"]
            etag_chunk_no_pairs.append(
                {"ETag": etag, "PartNumber": part_no}
            )  # maintain list of part no and ETag

        # Step 4 - Send a message to PyGrid informing about dataset upload complete!
        upload_response = client.datasets.perform_request(
            syft_msg=UploadDataCompleteMessage,
            content={
                "upload_id": upload_response.payload.upload_id,
                "filename": f"{dataset_name}/{asset_name}",
                "parts": etag_chunk_no_pairs,
            },
        )

        # Step 5 - Create a proxy dataset for the uploaded data.
        # Retrieve fully qualified name to  use for pointer creation.
        obj_public_kwargs = getattr(data, "proxy_public_kwargs", {})
        data_fqn = str(get_fully_qualified_name(data))
        data_dtype = str(type(data))

        # relative
        from ...tensor import Tensor
        from ...tensor.autodp.phi_tensor import PhiTensor as PT

        if isinstance(data, Tensor) and isinstance(data.child, PT):
            data_fqn = str(get_fully_qualified_name(data.child))
            child_kwargs = getattr(data.child, "proxy_public_kwargs", {})
            obj_public_kwargs.update(child_kwargs)

        proxy_data = ProxyDataset(
            asset_name=asset_name,
            dataset_name=dataset_name,
            node_id=client.id,
            dtype=data_dtype,
            fqn=data_fqn,
            shape=data.shape,
            obj_public_kwargs=obj_public_kwargs,
        )
    except (Exception, KeyboardInterrupt) as e:
        upload_id = upload_response.payload.upload_id
        abort_s3_object_upload(
            client=client, upload_id=upload_id, asset_name=asset_name
        )
        raise e

    return proxy_data


def get_s3_client(settings: BaseSettings) -> "boto3.client.S3":
    try:
        s3_endpoint = settings.S3_ENDPOINT
        s3_port = settings.S3_PORT
        s3_grid_url = GridURL(host_or_ip=s3_endpoint, port=s3_port)
        return boto3.client(
            "s3",
            endpoint_url=s3_grid_url.url,
            aws_access_key_id=settings.S3_ROOT_USER,
            aws_secret_access_key=settings.S3_ROOT_PWD,
            config=Config(signature_version="s3v4"),
            region_name=settings.S3_REGION,
        )
    except Exception as e:
        print(f"Failed to create S3 Client with {s3_endpoint} {s3_port} {s3_grid_url}")
        raise e


def check_send_to_blob_storage(obj: Any, use_blob_storage: bool = False) -> bool:
    """Check if the data needs to be send to Seaweed storage depending upon its size and type.
    Args:
        obj (Any): Data to be stored to Seaweed.
        use_blob_storage (bool, optional): Explicit flag to send the data to blob storage. Defaults to False.

    Returns:
        bool: Returns True if obj can be stored in blob store else returns False.
    """
    # relative
    from ...tensor import Tensor
    from ...tensor.autodp.phi_tensor import PhiTensor as PT

    if use_blob_storage and (
        isinstance(obj, (PT, Tensor)) or size_mb(obj) > MIN_BLOB_UPLOAD_SIZE_MB
    ):
        return True
    return False
